/*
 * Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License version 2.1 as published by the Free Software
 * Foundation.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301  USA
 */
package pl.chyla.wyjec;


/**
 * Configuration data for components of Wyjec language implementation. In defines mainly limits
 * of various memory resources.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class Config {

    /**
     * Maximum number of required parameters that may be declared by a function via its
     * signature.
     */
    public static final int FUNC_REQ_PARAMS_COUNT_LIMIT = 9;

    /**
     * Maximum length of function name supported by this Wyjec implementation. This limit
     * concerns only "real" functions, i.e. implemented as {@link pl.chyla.wyjec.core.Function}
     * instances, it doesn't limit the length of strings supported by
     * {@link pl.chyla.wyjec.core.IConstantsParser} instances.
     */
    public static final int FUNC_NAME_LENGTH_LIMIT = 9;

    /**
     * Default value for {@link #stringLengthLimit} property.
     */
    private static final int DEFAULT_STRING_LENGTH_LIMIT = 128;

    /**
     * Default value for {@link #funcActualParamsSoftLimit} property.
     */
    private static final int DEFAULT_FUNC_ACTUAL_PARAMS_SOFT_LIMIT = 64;

    /**
     * Default value for {@link #funcDepthSoftLimit} property.
     */
    private static final int DEFAULT_FUNC_DEPTH_SOFT_LIMIT = 16;

    /**
     * Default value for {@link #tmpValuesCountSoftLimit} property.
     */
    private static final int DEFAULT_TMP_VALUES_COUNT_SOFT_LIMIT = 128;

    /**
     * Default value for {@link #exprElementsCountSoftLimit} property.
     */
    private static final int DEFAULT_EXPR_ELEMENTS_COUNT_SOFT_LIMIT = 128;

    /**
     * Default value for {@link #stringConvFuncName} property.
     */
    private static final String DEFAULT_STRING_CONV_FUNC_NAME = "str";

    /**
     * Default value for {@link #stringConcatFuncName} property.
     */
    private static final String DEFAULT_STRING_CONCAT_FUNC_NAME = "+";

    /**
     * Maximum length of strings used for evaluation of expressions.
     */
    private int stringLengthLimit = DEFAULT_STRING_LENGTH_LIMIT;

    /**
     * Soft limit for number of actual parameters that may be passed to a function.
     */
    private int funcActualParamsSoftLimit = DEFAULT_FUNC_ACTUAL_PARAMS_SOFT_LIMIT;

    /**
     * Soft limit for depth of nested function references in evaluated Wyjec expressions.
     */
    private int funcDepthSoftLimit = DEFAULT_FUNC_DEPTH_SOFT_LIMIT;

    /**
     * Soft limit for number of temporary values used during evaluating of Wyjec expressions.
     */
    private int tmpValuesCountSoftLimit = DEFAULT_TMP_VALUES_COUNT_SOFT_LIMIT;

    /**
     * Soft limit for number of elements (literals, function references) in Wyjec expressions.
     */
    private int exprElementsCountSoftLimit = DEFAULT_EXPR_ELEMENTS_COUNT_SOFT_LIMIT;

    /**
     * Name of string conversion function to use by the parser
     */
    private String stringConvFuncName = DEFAULT_STRING_CONV_FUNC_NAME;

    /**
     * Name of string concatenation function to use by the parser
     */
    private String stringConcatFuncName = DEFAULT_STRING_CONCAT_FUNC_NAME;

    /**
     * Empty default constructor
     */
    public Config() { }

    /**
     * Returns new <code>Config</code> instance with default configuration data.
     *
     * @return new default <code>Config</code> instance
     */
    public static Config getDefaultConfig() {
        return new Config();
    }

    /**
     * Returns new <code>Config</code> instance with all limits affecting the amount of pooled
     * objects set to zero.
     *
     * @return new minimalistic <code>Config</code> instance
     */
    public static Config getZeroMemLimitsConfig() {
        Config conf = new Config();
        conf.setExprElementsCountSoftLimit(0);
        conf.setFuncActualParamsSoftLimit(0);
        conf.setFuncDepthSoftLimit(0);
        conf.setTmpValuesCountSoftLimit(0);
        return conf;
    }

    /**
     * Returns maximum length of strings used for evaluation of expressions. It concerns both
     * string representing final value of expression evaluation and and temporary strings
     * representing results of evaluations of nested functions.
     *
     * @return maximum length of strings used for evaluation of expressions
     * @see #setStringLengthLimit(int)
     */
    public int getStringLengthLimit() {
        return stringLengthLimit;
    }

    /**
     * Sets maximum length of strings used for evaluation of expressions.
     *
     * @param value new maximum length of strings used for evaluation of expressions
     * @see #getStringLengthLimit()
     */
    public void setStringLengthLimit(int value) {
        stringLengthLimit = value;
    }

    /**
     * Returns soft limit for number of actual parameters that may be passed to a function in
     * Wyjec expressions. Using more actual parameters may require dynamic memory allocation and
     * thus may be less effective.
     *
     * @return soft limit for number of actual parameters
     * @see #setFuncActualParamsSoftLimit(int)
     */
    public int getFuncActualParamsSoftLimit() {
        return funcActualParamsSoftLimit;
    }

    /**
     * Sets soft limit for number of actual parameters that may be passed to a function in Wyjec
     * expressions.
     *
     * @param value new soft limit for number of actual parameters
     * @see #getFuncActualParamsSoftLimit()
     */
    public void setFuncActualParamsSoftLimit(int value) {
        funcActualParamsSoftLimit = value;
    }

    /**
     * Returns soft limit for depth of nested function references in evaluated Wyjec
     * expressions. Evaluation of expressions with deeper nested function references may require
     * dynamic memory allocation and thus may be less effective.
     *
     * @return soft limit for depth of nested function references
     * @see #setFuncDepthSoftLimit(int)
     */
    public int getFuncDepthSoftLimit() {
        return funcDepthSoftLimit;
    }

    /**
     * Sets soft limit for depth of nested function references in evaluated Wyjec expressions.
     *
     * @param value new soft limit for depth of nested function references
     * @see #getFuncDepthSoftLimit()
     */
    public void setFuncDepthSoftLimit(int value) {
        funcDepthSoftLimit = value;
    }

    /**
     * Returns soft limit for number of temporary values used during evaluating of Wyjec
     * expressions. Evaluation of expressions that require more temporary values may require
     * dynamic memory allocation and thus may be less effective.
     *
     * @return soft limit for number of temporary values used during expression evaluation
     * @see #setTmpValuesCountSoftLimit(int)
     */
    public int getTmpValuesCountSoftLimit() {
        return tmpValuesCountSoftLimit;
    }

    /**
     * Sets soft limit for number of temporary values used during evaluating of Wyjec
     * expressions.
     *
     * @param value
     *     new soft limit for number of temporary values used during expression evaluation
     * @see #getTmpValuesCountSoftLimit()
     */
    public void setTmpValuesCountSoftLimit(int value) {
        tmpValuesCountSoftLimit = value;
    }

    /**
     * Returns soft limit for number of elements (literals, function references) in Wyjec
     * expressions. Parsing of more complex Wyjec expressions may require dynamic memory
     * allocation and thus may be less effective.
     *
     * @return soft limit for number of elements in expressions
     */
    public int getExprElementsCountSoftLimit() {
        return exprElementsCountSoftLimit;
    }

    /**
     * Sets soft limit for number of elements (literals, function references) in Wyjec
     * expressions.
     *
     * @param value new soft limit for number of elements expressions
     */
    public void setExprElementsCountSoftLimit(int value) {
        exprElementsCountSoftLimit = value;
    }

    /**
     * Returns the name of string conversion function to be used by Wyjec parser. Typically this
     * name should be overloaded to offer string conversion functions for various types. Every
     * function with this name should accept single parameter and return value of
     * {@link pl.chyla.wyjec.core.Type#STRING} type.
     *
     * @return name of string conversion function
     * @see #setStringConvFuncName(String)
     */
    public String getStringConvFuncName() {
        return stringConvFuncName;
    }

    /**
     * Sets the name of string conversion function to be used by Wyjec parser.
     *
     * @param funcName new name of string conversion function to use
     * @see #getStringConvFuncName()
     */
    public void setStringConvFuncName(String funcName) {
        stringConvFuncName = funcName;
    }

    /**
     * Returns the name of string concatenation function to be used by Wyjec parser. Function
     * with this name should accept any positive number of string parameters and return single
     * string value. It's possible to overload this name to use different functions for handling
     * different number of parameters, but this approach is not recommended.
     *
     * @return name of string concatenation function
     * @see #setStringConcatFuncName(String)
     */
    public String getStringConcatFuncName() {
        return stringConcatFuncName;
    }

    /**
     * Sets the name of string concatenation function to be used by Wyjec parser.
     *
     * @param funcName new name of string concatenation function to use
     * @see #getStringConcatFuncName()
     */
    public void setStringConcatFuncName(String funcName) {
        stringConcatFuncName = funcName;
    }

}
