/*
 * Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License version 2.1 as published by the Free Software
 * Foundation.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301  USA
 */
package pl.chyla.wyjec.apphelpers;

import pl.chyla.wyjec.Config;
import pl.chyla.wyjec.apphelpers.FixedSizeLRUCache.CacheEntry;
import pl.chyla.wyjec.core.Function;
import pl.chyla.wyjec.core.IConstantsParser;
import pl.chyla.wyjec.parser.ExprTreePool;
import pl.chyla.wyjec.parser.ExpressionTree;
import pl.chyla.wyjec.parser.InvalidInputException;
import pl.chyla.wyjec.parser.Parser;
import pl.chyla.wyjec.stdconsts.StandardConstantsParser;
import pl.chyla.wyjec.stdfuncs.StandardFunctions;
import pl.chyla.wyjec.utils.IObjectsFactory;
import pl.chyla.wyjec.utils.ObjectsPool;
import pl.chyla.wyjec.utils.RawString;


/**
 * Wrapper class for standard Wyjec {@link Parser parser} that caches parsed expression trees.
 * The cache keeps expression trees for some number of recently used expressions. The number of
 * cached expressions is limited be the size of the cache defined in its constructor.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class CachingParser {

    /**
     * Wyjec parser to use for real processing.
     */
    private final Parser parser;

    /**
     * Pool that provides expression tree elements for all cached expression trees.
     */
    private final ExprTreePool cachedExprTreesPool;

    /**
     * Pool of {@link ExpressionDescriptor} instances.
     */
    private final ObjectsPool exprDescsPool;

    /**
     * LRU cache that maps expression strings to {@link ExpressionDescriptor} instances.
     */
    private final FixedSizeLRUCache exprDescsCache;

    /**
     * Temporary {@link RawString} instance used to pass an expression to the real parser. This
     * field is used to avoid creating the instance dynamically.
     */
    private final RawString rawStrExpression;

    /**
     * Constructs <code>CachingParser</code> instance with support for standard set of Wyjec
     * functions and constants.
     *
     * @param cacheSize
     *     the number of recently used expressions that the cache will keep in parsed form (as
     *     expression trees)
     * @param inputBufferSize
     *     the size of character buffer used to store expressions before parsing them; if the
     *     length of an expression string passed to the parser is greater than this size, the
     *     parses will have to allocate temporary buffer for parsing an expression
     */
    public CachingParser(int cacheSize, int inputBufferSize) {
        this(
            Config.getDefaultConfig(),
            StandardFunctions.getFunctions(),
            StandardConstantsParser.getInstance(), cacheSize, inputBufferSize);
    }

    /**
     * Constructs <code>CachingParser</code> instance using given configuration, functions and
     * constants parser.
     *
     * @param config
     *     configuration to use during parsing and analysing of expressions by this parser
     * @param funcs array of functions to use during semantic analysis of expressions
     * @param constsParser constants parser to use during semantic analysis of expressions
     * @param cacheSize
     *     the number of recently used expressions that the cache will keep in parsed form (as
     *     expression trees)
     * @param inputBufferSize
     *     the size of character buffer used to store expressions before parsing them; if the
     *     length of an expression string passed to the parser is greater than this size, the
     *     parses will have to allocate temporary buffer for parsing an expression
     */
    public CachingParser(
        Config config, Function[] funcs, IConstantsParser constsParser, int cacheSize,
        int inputBufferSize) {
        if (cacheSize <= 0) {
            throw new IllegalArgumentException(
                "Cache size must be positive (" + cacheSize + " given)");
        }
        if (inputBufferSize < 0) {
            throw new IllegalArgumentException(
                "Input buffer size must not be negative (" + inputBufferSize + " given)");
        }
        parser = new Parser(config, funcs, constsParser);
        final int bufSize = inputBufferSize;
        exprDescsPool = new ObjectsPool(new IObjectsFactory() {
            public Object createObject() {
                return new ExpressionDescriptor(bufSize);
            }
        }, cacheSize);
        cachedExprTreesPool = new ExprTreePool(cacheSize, config);
        exprDescsCache = new FixedSizeLRUCache(cacheSize);
        rawStrExpression = new RawString();
    }

    /**
     * If given expression string already sits in the cache, this method returns cached tree
     * representation of the expression. Otherwise it parses the expression, stores the parse
     * tree in the cache and then returns it.
     * <p>
     * Note: returned expression tree will be build from objects got from the parser's internal
     * objects pool. These objects may be reused every time this method is called, so returned
     * expression tree will be valid only until the next call of this method.
     *
     * @param expression string containing expression to parse
     * @return tree corresponding to expression contained in the string
     * @throws InvalidInputException in case of syntax error in given input or a semantic error
     * @see Parser#parse(ReadableRawString)
     */
    public ExpressionTree parse(String expression) throws InvalidInputException {
        ExpressionDescriptor exprDesc = (ExpressionDescriptor) exprDescsCache.get(expression);
        if (exprDesc == null) {
            exprDesc = (ExpressionDescriptor) exprDescsPool.getObject();
            if (expression.length() <= exprDesc.getExpressionStorage().length) {
                char[] storage = exprDesc.getExpressionStorage();
                expression.getChars(0, expression.length(), storage, 0);
                rawStrExpression.set(storage, 0, expression.length());
            } else {
                rawStrExpression.set(expression.toCharArray());
            }
            ExpressionTree exprTree = parser.parse(rawStrExpression);
            if (!exprDescsCache.hasFreeSpace()) {
                CacheEntry oldCacheEntry = exprDescsCache.removeLeastRecentlyUsed();
                ExpressionDescriptor oldExprDesc =
                    (ExpressionDescriptor) oldCacheEntry.getValue();
                oldExprDesc.getExprTree().deepRelease(cachedExprTreesPool);
                exprDescsPool.releaseObject(oldExprDesc);
            }
            exprDesc.set(exprTree.makeDeepCopy(cachedExprTreesPool));
            exprDescsCache.add(expression, exprDesc);
        }
        return exprDesc.getExprTree();
    }


    /**
     * Object encapsulating single expression tree and character data that it refers to. It's
     * used as target object stored in LRU cache that identifies objects by string
     * representations of expressions.
     *
     * @author Zbigniew Chyla (zbigniew@chyla.pl)
     */
    private final class ExpressionDescriptor {

        /**
         * Buffer containing characters of expression which parsed form is stored in
         * {@link #exprTree}
         */
        private final char[] expressionStorage;

        /**
         * Embedded expression tree.
         */
        private ExpressionTree exprTree;

        /**
         * Constructs <code>ExpressionDescriptor</code> instance using given size for internal
         * character buffer.
         *
         * @param bufferSize size of the internal buffer
         */
        private ExpressionDescriptor(int bufferSize) {
            expressionStorage = new char[bufferSize];
        }

        /**
         * Sets embedded expression tree to given one.
         *
         * @param newExprTree expression tree to store inside the object
         */
        void set(ExpressionTree newExprTree) {
            exprTree = newExprTree;
        }

        /**
         * Returns embedded expression tree.
         *
         * @return expression tree stored inside the object
         */
        ExpressionTree getExprTree() {
            return exprTree;
        }

        /**
         * Returns internal character buffer that should be used to store characters of
         * expression before parsing it and passing its parsed form (as {@link ExpressionTree}
         * instance) to {@link #set(ExpressionTree)} method.
         *
         * @return internal buffer for storing characters of an expression
         */
        char[] getExpressionStorage() {
            return expressionStorage;
        }

    }

}
