/*
 * Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License version 2.1 as published by the Free Software
 * Foundation.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301  USA
 */
package pl.chyla.wyjec.core;

import pl.chyla.wyjec.Debugging;
import pl.chyla.wyjec.core.optparameters.BooleanOptParameters;
import pl.chyla.wyjec.core.optparameters.DateOptParameters;
import pl.chyla.wyjec.core.optparameters.FloatOptParameters;
import pl.chyla.wyjec.core.optparameters.IntegerOptParameters;
import pl.chyla.wyjec.core.optparameters.StringOptParameters;
import pl.chyla.wyjec.core.parameters.BooleanParameter;
import pl.chyla.wyjec.core.parameters.DateParameter;
import pl.chyla.wyjec.core.parameters.FloatParameter;
import pl.chyla.wyjec.core.parameters.IntegerParameter;
import pl.chyla.wyjec.core.parameters.StringParameter;
import pl.chyla.wyjec.utils.ReadableRawString;


/**
 * Abstract base class for all functions used in Wyjec expressions.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public abstract class Function {

    /**
     * Current evaluation context of the function.
     */
    private IEvaluationContext curEvalContext;

    /**
     * Evaluates this function's value for actual parameters got from given evaluation context
     * and stores the result in {@link Value} object it provides.
     *
     * @param evalContext
     *     evaluation context to use for accessing actual parameters and the result during this
     *     evaluation
     */
    public final void evaluate(IEvaluationContext evalContext) {
        // save the old evaluation context before evaluation and restore it afterwards, so the
        // evaluation method may be reentrant
        IEvaluationContext savedEvalContext = curEvalContext;
        curEvalContext = evalContext;
        typedEvaluate(evalContext.getResult());
        curEvalContext = savedEvalContext;
    }

    /**
     * Helper method providing access to {@link Value} instance containing value of specified
     * parameter.
     *
     * @param param function's parameter to provide value for
     * @return {@link Value} instance containing value of requested parameter
     */
    private Value getParameterValue(Parameter param) {
        int actParamIdx = param.getIndex();
        return curEvalContext.getParameter(actParamIdx);
    }

    /**
     * Helper method providing access to {@link Value} instance containing value of specified
     * optional parameter.
     *
     * @param optParams object grouping this function's optional parameters
     * @param index index of optional parameter to get value for
     * @return {@link Value} instance containing value of requested parameter
     */
    private Value getParameterValue(OptParameters optParams, int index) {
        int actParamIdx = optParams.getStartIndex() + index;
        if (Debugging.DEBUG_ENABLED) {
            Debugging.assume(
                actParamIdx < curEvalContext.getParametersCount(),
                "Given index of optional parameter (" + index + ") is out of valid range"
                + "<0, " + curEvalContext.getParametersCount() + ")");
        }
        return curEvalContext.getParameter(actParamIdx);
    }

    /**
     * Returns current value of this function's integer parameter specified by given
     * {@link IntegerParameter} object.
     * <p>
     * Method may be called only during evaluation of this function's result, calling it
     * at any other time will yield undefined value or
     * <code>ArrayIndexOutOfBoundsException</code>.
     *
     * @param param
     *     function's parameter to provide value for; it must be one of objects passed to
     *     constructor of signature ({@link FunctionSignature}) of this function
     * @return value of given parameter of type integer
     */
    protected final long getParameter(IntegerParameter param) {
        return getParameterValue(param).getInteger();
    }

    /**
     * Returns current value of this function's boolean parameter specified by given
     * {@link BooleanParameter} object.
     * <p>
     * Method may be called only during evaluation of this function's result, calling it
     * at any other time will yield undefined value or
     * <code>ArrayIndexOutOfBoundsException</code>.
     *
     * @param param
     *     function's parameter to provide value for; it must be one of objects passed to
     *     constructor of signature ({@link FunctionSignature}) of this function
     * @return value of given parameter of type boolean
     */
    protected final boolean getParameter(BooleanParameter param) {
        return getParameterValue(param).getBoolean();
    }

    /**
     * Returns current value of this function's string parameter specified by given
     * {@link StringParameter} object.
     * <p>
     * Method may be called only during evaluation of this function's result, calling it
     * at any other time will yield undefined value or
     * <code>ArrayIndexOutOfBoundsException</code>.
     *
     * @param param
     *     function's parameter to provide value for; it must be one of objects passed to
     *     constructor of signature ({@link FunctionSignature}) of this function
     * @return value of given parameter of type string
     */
    protected final ReadableRawString getParameter(StringParameter param) {
        return getParameterValue(param).getString();
    }

    /**
     * Returns current value of this function's float parameter specified by given
     * {@link FloatParameter} object.
     * <p>
     * Method may be called only during evaluation of this function's result, calling it
     * at any other time will yield undefined value or
     * <code>ArrayIndexOutOfBoundsException</code>.
     *
     * @param param
     *     function's parameter to provide value for; it must be one of objects passed to
     *     constructor of signature ({@link FunctionSignature}) of this function
     * @return value of given parameter of type float
     */
    protected final double getParameter(FloatParameter param) {
        return getParameterValue(param).getFloat();
    }

    /**
     * Returns current value of this function's date parameter specified by given
     * {@link DateParameter} object.
     * <p>
     * Method may be called only during evaluation of this function's result, calling it
     * at any other time will yield undefined value or
     * <code>ArrayIndexOutOfBoundsException</code>.
     *
     * @param param
     *     function's parameter to provide value for; it must be one of objects passed to
     *     constructor of signature ({@link FunctionSignature}) of this function
     * @return value of given parameter of type date
     */
    protected final long getParameter(DateParameter param) {
        return getParameterValue(param).getDate();
    }

    /**
     * Returns current value of this function's integer optional parameter specified
     * by given index.
     * <p>
     * Method may be called only during evaluation of this function's result, calling it at any
     * other time has undefined behaviour.
     *
     * @param optParams
     *     object grouping this function's optional parameters; it must be an object passed to
     *     constructor of signature ({@link FunctionSignature}) of this function
     * @param index
     *     index of optional parameter to get value for; indices of optional parameters start
     *     from 0 (numbering of optional parameters is independent of number of required
     *     parameters); it should be nonnegative number less than
     *     {@link #getOptParametersCount(OptParameters)}, for other values the behaviour of the
     *     method is undefined
     * @return value of optional parameter with specified index
     */
    protected final long getParameter(IntegerOptParameters optParams, int index) {
        return getParameterValue(optParams, index).getInteger();
    }

    /**
     * Returns current value of this function's boolean optional parameter specified
     * by given index.
     * <p>
     * Method may be called only during evaluation of this function's result, calling it at any
     * other time has undefined behaviour.
     *
     * @param optParams
     *     object grouping this function's optional parameters; it must be an object passed to
     *     constructor of signature ({@link FunctionSignature}) of this function
     * @param index
     *     index of optional parameter to get value for; indices of optional parameters start
     *     from 0 (numbering of optional parameters is independent of number of required
     *     parameters); it should be nonnegative number less than
     *     {@link #getOptParametersCount(OptParameters)}, for other values the behaviour of the
     *     method is undefined
     * @return value of optional parameter with specified index
     */
    protected final boolean getParameter(BooleanOptParameters optParams, int index) {
        return getParameterValue(optParams, index).getBoolean();
    }


    /**
     * Returns current value of this function's string optional parameter specified
     * by given index.
     * <p>
     * Method may be called only during evaluation of this function's result, calling it at any
     * other time has undefined behaviour.
     *
     * @param optParams
     *     object grouping this function's optional parameters; it must be an object passed to
     *     constructor of signature ({@link FunctionSignature}) of this function
     * @param index
     *     index of optional parameter to get value for; indices of optional parameters start
     *     from 0 (numbering of optional parameters is independent of number of required
     *     parameters); it should be nonnegative number less than
     *     {@link #getOptParametersCount(OptParameters)}, for other values the behaviour of the
     *     method is undefined
     * @return value of optional parameter with specified index
     */
    protected final ReadableRawString getParameter(StringOptParameters optParams, int index) {
        return getParameterValue(optParams, index).getString();
    }

    /**
     * Returns current value of this function's float optional parameter specified
     * by given index.
     * <p>
     * Method may be called only during evaluation of this function's result, calling it at any
     * other time has undefined behaviour.
     *
     * @param optParams
     *     object grouping this function's optional parameters; it must be an object passed to
     *     constructor of signature ({@link FunctionSignature}) of this function
     * @param index
     *     index of optional parameter to get value for; indices of optional parameters start
     *     from 0 (numbering of optional parameters is independent of number of required
     *     parameters); it should be nonnegative number less than
     *     {@link #getOptParametersCount(OptParameters)}, for other values the behaviour of the
     *     method is undefined
     * @return value of optional parameter with specified index
     */
    protected final double getParameter(FloatOptParameters optParams, int index) {
        return getParameterValue(optParams, index).getFloat();
    }

    /**
     * Returns current value of this function's date optional parameter specified
     * by given index.
     * <p>
     * Method may be called only during evaluation of this function's result, calling it at any
     * other time has undefined behaviour.
     *
     * @param optParams
     *     object grouping this function's optional parameters; it must be an object passed to
     *     constructor of signature ({@link FunctionSignature}) of this function
     * @param index
     *     index of optional parameter to get value for; indices of optional parameters start
     *     from 0 (numbering of optional parameters is independent of number of required
     *     parameters); it should be nonnegative number less than
     *     {@link #getOptParametersCount(OptParameters)}, for other values the behaviour of the
     *     method is undefined
     * @return value of optional parameter with specified index
     */
    protected final long getParameter(DateOptParameters optParams, int index) {
        return getParameterValue(optParams, index).getDate();
    }

    /**
     * Returns the number of currently available optional parameters (parameters passed to
     * this function after required parameters). If no optional parameters has been passed,
     * it returns <code>0</code>.
     * <p>
     * Method may be called only during evaluation of this function's result, calling it at any
     * other time has undefined behaviour.
     *
     * @param optParams
     *     object grouping this function's optional parameters; it must be an object passed to
     *     constructor of signature ({@link FunctionSignature}) of this function
     * @return the number of currently available optional parameters
     */
    protected final int getOptParametersCount(OptParameters optParams) {
        return curEvalContext.getParametersCount() - optParams.getStartIndex();
    }

    /**
     * Evaluates this function's value for currently set parameters (available via
     * <code>getParameter</code> methods) and stores the result in specified
     * {@link Value} object. Implementation of the method should set both value
     * and type of <code>result</code>.
     * <p>
     * Note: type of value set by this method must be the same as the type returned by
     * {@link #getResultType()}.
     *
     * @param result {@link Value} object to store function's result in
     */
    protected abstract void typedEvaluate(Value result);

    /**
     * Returns signature of this function, i.e. an object that describes function's parameters.
     * This method returns the same object for every call on the same instance.
     *
     * @return signature of this function
     * @see FunctionSignature
     */
    public abstract FunctionSignature getSignature();

    /**
     * Returns name used to identify this function in expressions (together with types of its
     * parameters in case of overloaded function's name). This method returns the same value for
     * every call on the same instance.
     *
     * @return a name of this function
     */
    public abstract String getName();

    /**
     * Returns type of values generated by this function. Returned value will be one of
     * <code>SimpleType.*</code>. This method must return the same type for all invocations on
     * any given {@link Function} instance.
     *
     * @return type of result of this function
     */
    public abstract Type getResultType();

}
