/*
 * Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License version 2.1 as published by the Free Software
 * Foundation.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301  USA
 */
package pl.chyla.wyjec.core;


/**
 * Declared parameter of a function. It's a part of function's signature representing type and
 * position of single parameter.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public abstract class Parameter {

    /**
     * Flag informing whether the parameter has been configured using {@link #configure(int)}
     * method.
     */
    private boolean isConfigured;

    /**
     * Nonnegative parameter's index on a list of declared required parameters contained in
     * function's signature the parameter is bound to.
     */
    private int index;

    /**
     * Constructs object representing function's parameter.
     * <p>
     * Note: Before the parameter may be used, it must be bound to a function's signature by
     * passing it to {@link FunctionSignature} constructor.
     */
    protected Parameter() { }

    /**
     * Configures this parameter by providing its index on a list of declared parameters. Every
     * <code>Parameter</code> must be configured before it can be used. Signature object that
     * <code>Parameter</code> instance is bound to is responsible for executing configuration
     * exactly once.
     *
     * @param paramIndex
     *     nonnegative position of the parameter in a list of declared parameters contained in
     *     the signature
     * @throws IllegalStateException in case the parameter is already configured
     * @throws IllegalAccessError in case given index is negative
     */
    final void configure(int paramIndex) {
        if (isConfigured) {
            throw new IllegalStateException("Parameter may be configured only once");
        }
        if (paramIndex < 0) {
            throw new IllegalArgumentException("Parameter's index must not be negative");
        }
        isConfigured = true;
        index = paramIndex;
    }

    /**
     * Returns the parameter's index on a list of parameters contained in function's
     * signature it's bound to
     * <p>
     * Note: Before the method can be used, the parameter must be configured (by binding it to
     * a signature).
     *
     * @return index of the parameter on a list of declared required parameters
     * @throws IllegalStateException in case the parameter hasn't been configured yet
     */
    public final int getIndex() {
        if (!isConfigured) {
            throw new IllegalStateException(
                "Parameter object must be configured before it can be used");
        }
        return index;
    }

    /**
     * Returns type of the parameter
     *
     * @return type of the parameter
     */
    public abstract Type getType();

}
