/*
 * Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License version 2.1 as published by the Free Software
 * Foundation.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301  USA
 */
package pl.chyla.wyjec.core;

import pl.chyla.wyjec.utils.IntUtils;


/**
 * Type of simple value used in expression or function's declaration.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class Type {

    /**
     * Number of bits required to represent encoded value of a type (as returned by
     * {@link #getEncodedType()}).
     */
    public static final int ENCODED_TYPE_BIT_SIZE = 3;

    /**
     * Bit mask covering all bits used in encoded value of a type (as returned by
     * {@link #getEncodedType()}).
     */
    public static final int ENCODED_TYPE_BIT_MASK = (1 << ENCODED_TYPE_BIT_SIZE) - 1;

    /**
     * Encoded form of type <code>NONE</code>
     */
    public static final int NONE_ENCODED = 0;

    /**
     * Encoded form of type <code>BOOLEAN</code>
     */
    public static final int BOOLEAN_ENCODED = 1;

    /**
     * Encoded form of type <code>INTEGER</code>
     */
    public static final int INTEGER_ENCODED = 2;

    /**
     * Encoded form of type <code>STRING</code>
     */
    public static final int STRING_ENCODED = 3;

    /**
     * Encoded form of type <code>FLOAT</code>
     */
    public static final int FLOAT_ENCODED = 4;

    /**
     * Encoded form of type <code>DATE</code>
     */
    public static final int DATE_ENCODED = 5;

    /**
     * Object representing type with no values, used when a type of a value is required but no
     * value exists.
     */
    public static final Type NONE = new Type(NONE_ENCODED, "<none>");

    /**
     * Boolean type with two values: true and false.
     */
    public static final Type BOOLEAN = new Type(BOOLEAN_ENCODED, "boolean");

    /**
     * Integer type with the same set of values as Java's <code>long<code> type.
     */
    public static final Type INTEGER = new Type(INTEGER_ENCODED, "integer");

    /**
     * String type with values consisting of characters from Unicode set. Length of values is
     * limited but this limit is not part of the type.
     */
    public static final Type STRING = new Type(STRING_ENCODED, "string");

    /**
     * Numeric type with set of values including subset of real numbers, represented using
     * floating point notation.
     */
    public static final Type FLOAT = new Type(FLOAT_ENCODED, "float");

    /**
     * Date type representing date and time.
     */
    public static final Type DATE = new Type(DATE_ENCODED, "date");

    /**
     * Mapping from encoded forms of types (table's indices) to <code>Type</code> objects
     */
    private static final Type[] DECODING_TABLE = {
        NONE, BOOLEAN, INTEGER, STRING, FLOAT, DATE};

    /**
     * Numeric value representing encoded form of the type.
     */
    private final int encodedType;

    /**
     * Returns human-readable and unique name of the type
     */
    private final String name;

    /**
     * Private constructor to disallow instantiating of the class from outside.
     *
     * @param encodedType encoded form of the type as <code>int</code> number
     * @param name a name of the type
     */
    private Type(int encodedType, String name) {
        this.encodedType = encodedType;
        this.name = name;
    }

    /**
     * Decodes type encoded to <code>int</code> value (as returned by
     * {@link #getEncodedType()}).
     *
     * @param encodedType type encoded as <code>int</code> value
     * @return type object that has encoded form equal to given number
     * @throws IllegalArgumentException in case given number is not valid encoded type
     */
    public static Type decodeType(int encodedType) {
        if (!IntUtils.isBetweenInclusive(encodedType, 0, DECODING_TABLE.length - 1)) {
            throw new IllegalArgumentException("Invalid encoded type: " + encodedType);
        }
        return DECODING_TABLE[encodedType];
    }

    /**
     * Returns human-readable and unique name of the type
     *
     * @return a name of the type
     */
    public String getName() {
        return name;
    }

    /**
     * Returns <code>int</code> value representing encoded form of the type. The number may be
     * decoded to original object with {@link #decodeType(int)}.
     *
     * @return <code>int</code> number representing the type
     */
    public int getEncodedType() {
        return encodedType;
    }

}
