/*
 * Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License version 2.1 as published by the Free Software
 * Foundation.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301  USA
 */
package pl.chyla.wyjec.parser;

import pl.chyla.wyjec.Config;
import pl.chyla.wyjec.core.Value;
import pl.chyla.wyjec.utils.IObjectsFactory;
import pl.chyla.wyjec.utils.ObjectsPool;


/**
 * Pool of objects used to build expression tree. It provides one or more
 * {@link ExpressionTree} instances and the proportional number of instances of other types
 * representing nodes in expression tree.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class ExprTreePool {

    /**
     * Pool of {@link ExpressionTree} objects used to build expression tree.
     */
    private final ObjectsPool expressionTreePool;

    /**
     * Pool of {@link FunctionNode} objects used to build expression tree.
     */
    private final ObjectsPool functionNodePool;

    /**
     * Pool of {@link ValueNode} objects used to build expression tree.
     */
    private final ObjectsPool valueNodePool;

    /**
     * Pool of {@link ParameterNode} objects used to build expression tree.
     */
    private final ObjectsPool parameterNodePool;

    /**
     * Pool of {@link Value} objects used to build expression tree.
     */
    private final ObjectsPool valuePool;

    /**
     * Constructs <code>ExprTreePool</code> instance with specified number of preallocated
     * {@link ExpressionTree} instances and instances of other types representing nodes in
     * expression tree, using number of elements specified in given {@link Config} object
     * multiplied by the number of {@link ExpressionTree} instances. It means that the number
     * of preallocated expression elements will be proportional to the number of
     * {@link ExpressionTree} instances.
     *
     * @param exprsCount number of {@link ExpressionTree} instances to preallocate
     * @param config
     *     {@link Config} object defining the number of other expression tree elements to
     *     preallocate
     */
    public ExprTreePool(int exprsCount, Config config) {
        expressionTreePool = new ObjectsPool(
            new IObjectsFactory() {
                public Object createObject() {
                    return new ExpressionTree();
                }
            }, exprsCount);
        functionNodePool = new ObjectsPool(
            new IObjectsFactory() {
                public Object createObject() {
                    return new FunctionNode();
                }
            }, exprsCount * config.getExprElementsCountSoftLimit());
        valueNodePool = new ObjectsPool(
            new IObjectsFactory() {
                public Object createObject() {
                    return new ValueNode();
                }
            }, exprsCount * config.getExprElementsCountSoftLimit());
        parameterNodePool = new ObjectsPool(
            new IObjectsFactory() {
                public Object createObject() {
                    return new ParameterNode();
                }
            }, exprsCount * config.getExprElementsCountSoftLimit());
        valuePool = new ObjectsPool(
            new IObjectsFactory() {
                public Object createObject() {
                    return new Value(0);
                }
            }, exprsCount * config.getExprElementsCountSoftLimit());
    }

    /**
     * Resets this pool to its initial state. After calling this method, any objects previously
     * acquired from this pool may be used only until the next call to one of the
     * <code>get*</code> methods. In most cases it simply means that after calling this method
     * using any previously acquired objects is forbidden.
     */
    public void reset() {
        expressionTreePool.reset();
        functionNodePool.reset();
        valueNodePool.reset();
        parameterNodePool.reset();
        valuePool.reset();
    }

    /**
     * Fetches new {@link ExpressionTree} instance from the pool. After calling {@link #reset()}
     * method this instance will be returned to the pool (together with all other ones) and
     * using it later will not be allowed.
     *
     * @return {@link ExpressionTree} instance acquired from this pool
     */
    ExpressionTree getExpressionTree() {
        return (ExpressionTree) expressionTreePool.getObject();
    }

    /**
     * Returns given {@link ExpressionTree} instance to the pool. The instance should be one of
     * instances previously acquired with {@link #getExpressionTree()} method from this pool.
     * After calling this method, using the instance given as a parameter is forbidden.
     *
     * @param expressionTree {@link ExpressionTree} instance to return to the pool
     */
    void releaseExpressionTree(ExpressionTree expressionTree) {
        expressionTreePool.releaseObject(expressionTree);
    }

    /**
     * Fetches new {@link FunctionNode} instance from the pool. After calling {@link #reset()}
     * method this instance will be returned to the pool (together with all other ones) and
     * using it later will not be allowed.
     *
     * @return {@link FunctionNode} instance acquired from this pool
     */
    FunctionNode getFunctionNode() {
        return (FunctionNode) functionNodePool.getObject();
    }

    /**
     * Returns given {@link FunctionNode} instance to the pool. The instance should be one of
     * instances previously acquired with {@link #getFunctionNode()} method from this pool.
     * After calling this method, using the instance given as a parameter is forbidden.
     *
     * @param functionNode {@link FunctionNode} instance to return to the pool
     */
    void releaseFunctionNode(FunctionNode functionNode) {
        functionNodePool.releaseObject(functionNode);
    }

    /**
     * Fetches new {@link ValueNode} instance from the pool. After calling {@link #reset()}
     * method this instance will be returned to the pool (together with all other ones) and
     * using it later will not be allowed.
     *
     * @return {@link ValueNode} instance acquired from this pool
     */
    ValueNode getValueNode() {
        return (ValueNode) valueNodePool.getObject();
    }

    /**
     * Returns given {@link ValueNode} instance to the pool. The instance should be one of
     * instances previously acquired with {@link #getValueNode()} method from this pool. After
     * calling this method, using the instance given as a parameter is forbidden.
     *
     * @param valueNode {@link ValueNode} instance to return to the pool
     */
    void releaseValueNode(ValueNode valueNode) {
        valueNodePool.releaseObject(valueNode);
    }

    /**
     * Fetches new {@link ParameterNode} instance from the pool. After calling {@link #reset()}
     * method this instance will be returned to the pool (together with all other ones) and
     * using it later will not be allowed.
     *
     * @return {@link ParameterNode} instance acquired from this pool
     */
    ParameterNode getParameterNode() {
        return (ParameterNode) parameterNodePool.getObject();
    }

    /**
     * Returns given {@link ParameterNode} instance to the pool. The instance should be one of
     * instances previously acquired with {@link #getParameterNode()} method from this pool.
     * After calling this method, using the instance given as a parameter is forbidden.
     *
     * @param parameterNode {@link ParameterNode} instance to return to the pool
     */
    void releaseParameterNode(ParameterNode parameterNode) {
        parameterNodePool.releaseObject(parameterNode);
    }

    /**
     * Fetches new {@link Value} instance from the pool. After calling {@link #reset()} method
     * this instance will be returned to the pool (together with all other ones) and using it
     * later will not be allowed.
     * <p>
     * Note: length of characters storage of {@link pl.chyla.wyjec.utils.RawString} value
     * contained in the returned instance will be equal to <code>0</code>, so it will not be
     * possible to append any characters to it. However, it will be possible to use some other
     * characters storage for string value, using {@link Value#setString(ReadableRawString)}
     * method.
     *
     * @return {@link Value} instance acquired from this pool
     */
    Value getValue() {
        return (Value) valuePool.getObject();
    }

    /**
     * Returns given {@link Value} instance to the pool. The instance should be one of instances
     * previously acquired with {@link #getValue()} method from this pool. After calling this
     * method, using the instance given as a parameter is forbidden.
     *
     * @param value {@link Value} instance to return to the pool
     */
    void releaseValue(Value value) {
        valuePool.releaseObject(value);
    }

}
