/*
 * Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License version 2.1 as published by the Free Software
 * Foundation.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301  USA
 */
package pl.chyla.wyjec.parser;


/**
 * Represents the whole expression tree, providing access to its top-level element.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class ExpressionTree {

    /**
     * Top-level atom of this expression tree.
     */
    private AtomNode topLevelAtom;

    /**
     * Constructs <code>ExpressionTree</code> instance.
     */
    ExpressionTree() { }

    /**
     * Initializes this <code>ExpressionTree</code> instance, setting its internal field.
     *
     * @param newTopAtom atom to use as top-level atom of this expression tree
     */
    void set(AtomNode newTopAtom) {
        topLevelAtom = newTopAtom;
    }

    /**
     * Makes deep copy of expression tree represented by this <code>ExpressionTree</code>
     * instance, using instances taken from given pool for copied nodes of the tree.
     *
     * @param pool
     *     pool used as the source of instances that will represent nodes of the copied tree
     * @return
     *     <code>ExpressionTree</code> instance taken from given pool and representing copy of
     *     this expression tree
     */
    public ExpressionTree makeDeepCopy(ExprTreePool pool) {
        ExpressionTree copy = pool.getExpressionTree();
        copy.topLevelAtom = topLevelAtom.makeDeepCopy(pool);
        return copy;
    }

    /**
     * Recursively returns all objects used to represent this expression tree (including this
     * <code>ExpressionTree</code> instance itself) to given pool. After calling this method
     * this tree should no longed be used.
     *
     * @param pool pool of expression tree objects that should be used to release this tree
     */
    public void deepRelease(ExprTreePool pool) {
        topLevelAtom.deepRelease(pool);
        pool.releaseExpressionTree(this);
    }

    /**
     * Returns reference to top-level atom of expression tree (typically it will be top-level
     * function of the expression).
     *
     * @return top-level atom of expression tree
     */
    public AtomNode getTopLevelAtom() {
        return topLevelAtom;
    }

}
