/*
 * Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License version 2.1 as published by the Free Software
 * Foundation.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301  USA
 */
package pl.chyla.wyjec.parser;

import pl.chyla.wyjec.lexer.Lexer;


/**
 * Exception thrown in case an input string of Wyjec parser is invalid because of syntax or
 * semantic error, e.g. the use of undefined function.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class InvalidInputException extends Exception {

    /**
     * Input string with Wyjec expression.
     */
    private final String inputString;

    /**
     * 0-based position of the error.
     */
    private final int errorPos;

    /**
     * Constructs new <code>InvalidInputException</code> instance that describes an error in
     * input string at start position of current token (both input string and token's position
     * will be got from given lexer instance).
     *
     * @param lexer lexer instance used during parsing
     * @param msg message to append to automatically generated part
     */
    InvalidInputException(Lexer lexer, String msg) {
        this(lexer, lexer.getTokenStartIndex(), msg);
    }

    /**
     * Constructs new <code>InvalidInputException</code> instance that describes an error in
     * input string at given position (input string will be got from given lexer instance).
     *
     * @param lexer lexer instance used during parsing
     * @param errorPos location of error in a string representing Wyjec expression
     * @param msg message to append to automatically generated part
     */
    InvalidInputException(Lexer lexer, int errorPos, String msg) {
        super(msg);
        this.inputString = lexer.getInputString().toString();
        this.errorPos = errorPos;
    }

    /**
     * Returns 0-based position in the input string at which the parser detected the error.
     *
     * @return position of the detected error
     */
    public int getErrorPos() {
        return errorPos;
    }

    /**
     * Returns an input string in which the parser detected the error.
     *
     * @return input string with erroneous expression
     */
    public String getInputString() {
        return inputString;
    }

}
