/*
 * Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License version 2.1 as published by the Free Software
 * Foundation.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301  USA
 */
package pl.chyla.wyjec.parser;


/**
 * Expression tree node representing single function parameter referencing to parameter atom and
 * (optionally) to other parameters of the same function.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class ParameterNode {

    /**
     * Atom corresponding to this function parameter
     */
    private AtomNode atom;

    /**
     * Node representing next parameter of the same function
     */
    private ParameterNode next;

    /**
     * Constructs <code>ParameterNode</code> instance.
     */
    ParameterNode() { }

    /**
     * Initializes this <code>ParameterNode</code> instance, setting its reference to
     * parameter's atom. This method should be called for the first parameter of a function,
     * after that either {@link #setLast()} should be called (indicating that this is the last
     * parameter) or {@link #setAfter(ParameterNode, AtomNode)} should be called on some other
     * instance with this instance given as "previous parameter".
     *
     * @param newAtom atom to be referenced by this parameter
     */
    void setFirst(AtomNode newAtom) {
        atom = newAtom;
    }

    /**
     * Initializes this <code>ParameterNode</code> instance, setting its reference to
     * parameter's atom, and sets this parameter as the consequent of parameter given as
     * <code>prevParam</code>. After calling this method either the same method should be called
     * on some other instance with this instance given as <code>prevParam</code>, or
     * {@link #setLast()} should be called indicating that this is the last parameter.
     *
     * @param prevParam parameter that is to preceed this one in the list of function parameters
     * @param newAtom atom to be referenced by this parameter
     */
    void setAfter(ParameterNode prevParam, AtomNode newAtom) {
        prevParam.next = this;
        atom = newAtom;
    }

    /**
     * Initializes this <code>ParameterNode</code> instance, by indicating that this is the
     * last parameter in the list of function parameters. Before calling this method either
     * {@link #setFirst(AtomNode)} method should be called (if this is the first parameter) or
     * {@link #setAfter(ParameterNode, AtomNode)} one (if this is consequent of some other
     * parameter).
     */
    void setLast() {
        next = null;
    }

    /**
     * Returns atom corresponding to this function parameter
     *
     * @return this parameter's atom
     */
    public AtomNode getAtom() {
        return atom;
    }

    /**
     * Returns expression tree node representing next parameter of the same function
     *
     * @return
     *     node of next parameter of the function or <code>null</code> if this node represents
     *     the last parameter of the function
     */
    public ParameterNode getNext() {
        return next;
    }

    /**
     * Makes deep copy of part of expression tree for which this parameter node is top-level
     * node, i.e. copies this node and all nodes referenced from it. Given pool will be used to
     * get instances for copies of this nodes and all referenced nodes of expression tree.
     *
     * @param pool
     *     pool used as the source of instances that will represent nodes of the copied part of
     *     expression tree
     * @return
     *     <code>ParameterNode</code> instance taken from given pool and representing copy of
     *     part of expression tree starting from this parameter node
     */
    ParameterNode makeDeepCopy(ExprTreePool pool) {
        ParameterNode copy = pool.getParameterNode();
        copy.atom = atom.makeDeepCopy(pool);
        if (next != null) {
            copy.next = next.makeDeepCopy(pool);
        } else {
            copy.next = null;
        }
        return copy;
    }

    /**
     * Recursively returns this node and all its inferior elements of expression tree to given
     * pool. This instance should be one of <code>ParameterNode</code> instances previously
     * acquired with {@link ExprTreePool#getParameterNode()} method from the same pool. After
     * calling this method this object should no longed be used.
     *
     * @param pool
     *     pool of expression tree objects that should be used to release this node and its
     *     children
     */
    void deepRelease(ExprTreePool pool) {
        atom.deepRelease(pool);
        if (next != null) {
            next.deepRelease(pool);
        }
        pool.releaseParameterNode(this);
    }

}
