/*
 * Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License version 2.1 as published by the Free Software
 * Foundation.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301  USA
 */
package pl.chyla.wyjec.parser;

import java.util.Arrays;
import pl.chyla.wyjec.Config;
import pl.chyla.wyjec.core.Function;
import pl.chyla.wyjec.core.Type;
import pl.chyla.wyjec.repo.FunctionsRepository;
import pl.chyla.wyjec.utils.RawString;
import pl.chyla.wyjec.utils.ReadableRawString;


/**
 * Factory for functions with special meaning for Wyjec parser. It provides special functions
 * by getting their names from given configuration object and searching them in provided
 * functions repository.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
final class SpecialFunctionsProvider {

    /**
     * Number of types of leading actual parameters that should be used to choose an overloaded
     * function. The remaining parameters may be ignored.
     */
    private static final int DISTINGUISHABLE_PARAMS_TYPES_COUNT =
        Config.FUNC_REQ_PARAMS_COUNT_LIMIT + 1;

    /**
     * Repository of functions used as source for special functions.
     */
    private final FunctionsRepository funcs;

    /**
     * Name of the string conversion function (based on given configuration)
     */
    private final ReadableRawString convFuncName;

    /**
     * Array of conversion functions indexed by encoded types of their parameters. This array
     * gets filled lazily, i.e. when conversion function with some source type is requested
     * for the first time.
     */
    private final Function[] convFuncsByType;

    /**
     * Name of the string concatenation function (based on given configuration)
     */
    private final ReadableRawString concatFuncName;

    /**
     * Array of conversion functions indexed by the number of accepted parameters (the same
     * function may appear mutiple times in the array if it has optional parameters). This array
     * gets filled lazily, i.e. when concatenation function with some number of accepted
     * parameters is requested for the first time.
     */
    private final Function[] concatFuncsByParamsCount;

    /**
     * Temporary value to avoid dynamic allocation of objects; array of types of actual
     * parameters for function.
     */
    private final Type[] tmpParamsTypes;

    /**
     * Constructs <code>SpecialFunctionsProvider</code> based on configuration stored in given
     * {@link Config} instance and providing functions from given functions repository.
     *
     * @param config configuration defining names of special functions
     * @param funcsRepo repository of functions to use as the source for special functions
     */
    SpecialFunctionsProvider(Config config, FunctionsRepository funcsRepo) {
        funcs = funcsRepo;
        convFuncName = new RawString(config.getStringConvFuncName());
        convFuncsByType = new Function[1 << Type.ENCODED_TYPE_BIT_SIZE];
        concatFuncName = new RawString(config.getStringConcatFuncName());
        concatFuncsByParamsCount = new Function[DISTINGUISHABLE_PARAMS_TYPES_COUNT + 1];
        tmpParamsTypes = new Type[DISTINGUISHABLE_PARAMS_TYPES_COUNT];
    }

    /**
     * Returns string conversion function for given source type, i.e. a function that accepts
     * single parameter of given type and returns value of type string.
     *
     * @param srcType type of parameter that is to be accepted by requested function
     * @return
     *     string conversion function accepting single parameter of given type and returning
     *     string value or <code>null</code> if no matching function can be found in the
     *     repository
     */
    Function getStrConversionFunc(Type srcType) {
        Function func = convFuncsByType[srcType.getEncodedType()];
        if (func == null) {
            tmpParamsTypes[0] = srcType;
            func = funcs.lookup(convFuncName, tmpParamsTypes, 1);
            if (func != null) {
                if (func.getResultType() == Type.STRING) {
                    convFuncsByType[srcType.getEncodedType()] = func;
                } else {
                    func = null;
                }
            }
        }
        return func;
    }

    /**
     * Returns string concatenation function (with name specified in the configuration) that
     * can accept given number of string parameters.
     *
     * @param paramsCount number of parameters that is to be accepted by requested function
     * @return
     *     string concatenation function accepting <code>paramsCount</code> string parameters
     *     and returning string value or <code>null</code> if no matching function can be found
     *     in the repository
     */
    Function getStrConcatenationFunc(int paramsCount) {
        int checkedParamsCount = Math.min(paramsCount, DISTINGUISHABLE_PARAMS_TYPES_COUNT);
        Function func = concatFuncsByParamsCount[checkedParamsCount];
        if (func == null) {
            Arrays.fill(tmpParamsTypes, 0, checkedParamsCount, Type.STRING);
            func = funcs.lookup(concatFuncName, tmpParamsTypes, checkedParamsCount);
            if (func != null) {
                if (func.getResultType() == Type.STRING) {
                    concatFuncsByParamsCount[checkedParamsCount] = func;
                } else {
                    func = null;
                }
            }
        }
        return func;
    }

}
