/*
 * Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License version 2.1 as published by the Free Software
 * Foundation.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301  USA
 */
package pl.chyla.wyjec.repo;

import java.util.Comparator;
import pl.chyla.wyjec.core.Function;
import pl.chyla.wyjec.core.FunctionSignature;
import pl.chyla.wyjec.core.Type;
import pl.chyla.wyjec.utils.LongUtils;


/**
 * Comparator used to compare functions by their names and (in case of functions with equal
 * names) by types of their parameters.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
final class FunctionComparator implements Comparator {

    /**
     * Instance of <code>FunctionNameComparator</code>, accessible from outside via
     * {@link #getInstance()}.
     */
    private static final FunctionComparator instance = new FunctionComparator();

    /**
     * Private (empty) constructor to disallow instantiating of the class from outside. Class'
     * instance should be accessed via {@link #getInstance()} getter.
     */
    private FunctionComparator() {
        // empty
    }

    /**
     * Returns singleton instance of <code>FunctionNameComparator</code>.
     *
     * @return instance of <code>FunctionNameComparator</code>
     */
    static FunctionComparator getInstance() {
        return instance;
    }

    /**
     * Compares two functions for order. For functions with different names it's an
     * equivalent for calling <code>compareTo</code> method on name of first function with
     * second functions' name as a parameter (comparision of <code>String</code> objects).
     * For functions with equal names it's an equivalent for calling
     * {@link LongUtils#compare(long, long)} on functions' parameters' types encoded as
     * <code>long</code> numbers.
     *
     * @param a first function used in comparision
     * @param b second function used in comparision
     * @return
     *     negative number, zero or positive number if the function is less, equal or greater
     *     than function passed as a parameter
     * @throws ClassCastException if some of arguments is not {@link Function} instance
     */
    public int compare(Object a, Object b) {
        Function aFunc = (Function) a;
        Function bFunc = (Function) b;
        int nameCmp = aFunc.getName().compareTo(bFunc.getName());
        if (nameCmp != 0) {
            return nameCmp;
        } else {
            FunctionSignature aSignature = aFunc.getSignature();
            FunctionSignature bSignature = bFunc.getSignature();
            Type[] aRequiredParamsTypes = aSignature.getRequiredParametersTypes();
            Type[] bRequiredParamsTypes = bSignature.getRequiredParametersTypes();
            long encParamsTypesA = EncodedTypeList.encodeTypeList(
                aRequiredParamsTypes, aRequiredParamsTypes.length,
                aSignature.getOptionalParametersType());
            long encParamsTypesB = EncodedTypeList.encodeTypeList(
                bRequiredParamsTypes, bRequiredParamsTypes.length,
                bSignature.getOptionalParametersType());
            return LongUtils.compare(encParamsTypesA, encParamsTypesB);
        }

    }

}
