/*
 * Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License version 2.1 as published by the Free Software
 * Foundation.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301  USA
 */
package pl.chyla.wyjec.repo;

import pl.chyla.wyjec.Config;
import pl.chyla.wyjec.Debugging;
import pl.chyla.wyjec.utils.IntUtils;
import pl.chyla.wyjec.utils.LongUtils;
import pl.chyla.wyjec.utils.ReadableRawString;


/**
 * Class responsible for encoding functions' names to numbers of type <code>long</long>.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
final class FunctionNameEncoder {

    /**
     * Number of used bits in characters of valid functions' names
     */
    private static final int VALID_CHAR_BIT_SIZE = 7;

    /**
     * Bit mask covering all used bits in characters of valid functions' names
     */
    private static final int VALID_CHAR_BIT_MASK = (1 << VALID_CHAR_BIT_SIZE) - 1;

    /**
     * Verifies whether limits defined by {@link Config} class are not too loose.
     */
    static {
        if (Debugging.DEBUG_ENABLED) {
            int maxFunctionNameLength = (LongUtils.LONG_BITS_COUNT - 1) / VALID_CHAR_BIT_SIZE;
            Debugging.assume(
                Config.FUNC_NAME_LENGTH_LIMIT <= maxFunctionNameLength,
                "Limits.MAX_FUNCTION_NAME_LENGTH must not be greater than "
                + maxFunctionNameLength);
        }
    }

    /**
     * Private empty constructor to disallow instantiating of the class.
     */
    private FunctionNameEncoder() { }

    /**
     * Encodes function's name to simple number of type <code>long</long>.
     *
     * @param name a name to encode
     * @return <code>long</code> number representing encoded form of given name
     */
    static long encodeName(ReadableRawString name) {
        long encName = 0;
        char[] chars = name.getStorage();
        int nameStart = name.getStartPos();
        int nameEnd = name.getEndPos();
        int charShift = VALID_CHAR_BIT_SIZE * (Config.FUNC_NAME_LENGTH_LIMIT - 1);
        for (int i = nameStart; i < nameEnd; i++) {
            encName |= ((long) chars[i]) << charShift;
            charShift -= VALID_CHAR_BIT_SIZE;
        }
        return encName;
    }

    /**
     * Verifies whether given string is valid as function's name. Function's name must be a
     * string of ASCII characters with length between 1 and MAX_NAME_LENGTH (inclusive).
     *
     * @param name a name to check
     * @return <code>true</code> if given name is valid, <code>false</code> otherwise
     */
    static boolean isValidName(ReadableRawString name) {
        int len = name.getLength();
        if (!IntUtils.isBetweenInclusive(len, 1, Config.FUNC_NAME_LENGTH_LIMIT)) {
            return false;
        }
        char[] chars = name.getStorage();
        char charsBitSum = 0;
        int endPos = name.getEndPos();
        for (int i = name.getStartPos(); i < endPos; i++) {
            charsBitSum |= chars[i];
        }
        return (charsBitSum & VALID_CHAR_BIT_MASK) == charsBitSum;
    }

}
