/*
 * Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License version 2.1 as published by the Free Software
 * Foundation.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301  USA
 */
package pl.chyla.wyjec.repo;

import pl.chyla.wyjec.Config;
import pl.chyla.wyjec.core.Function;
import pl.chyla.wyjec.core.FunctionSignature;
import pl.chyla.wyjec.core.Type;


/**
 * Group of functions with equal names ("overloading" single name) that can be searched using a
 * sequence of types of actual parameters to be accepted by one of the functions.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
abstract class FunctionsGroup {

    /**
     * Creates new instance of <code>FunctionGroup</code> representing a group that contains
     * given functions. The actual class (subclass of <code>FunctionGroup</code>) of the group
     * depends on the number of functions it should contain.
     *
     * @param funcs array of functions to put in the group
     * @return new <code>FunctionGroup</code> instance
     * @throws IllegalArgumentException
     *     in case functions in the array doesn't obey rules for function's group: it must not
     *     be empty, all functions within group must have the same name, all functions must have
     *     disjunctive parameters' types
     */
    static FunctionsGroup createGroup(Function[] funcs) {
        checkGroup(funcs);
        if (funcs.length == 1) {
            return new SingletonGroup(funcs[0]);
        } else {
            return new FullGroup(funcs);
        }
    }

    /**
     * Returns name of the group, defined as common name of all functions it contains.
     *
     * @return name of the group
     */
    abstract String getName();

    /**
     * Searches the group for a function that can accept parameters with given types.
     *
     * @param paramsTypes
     *     array of types of actual parameters; its length may exceed actual number of
     *     parameters (specified by <code>paramsCount</code> parameter), in this case only
     *     array's components with indices less than <code>paramsCount</code> will be used.
     * @param paramsCount number of actual parameters (positive or zero)
     * @return
     *     {@link Function} instance that can accept actual parameters with given types or
     *     <code>null</code> if no such function exists
     */
    final Function lookupFunction(Type[] paramsTypes, int paramsCount) {
        long actEncParamsTypes;
        if (paramsCount <= Config.FUNC_REQ_PARAMS_COUNT_LIMIT) {
            actEncParamsTypes = EncodedTypeList.encodeTypeList(
                paramsTypes, paramsCount, Type.NONE);
        } else {
            Type optParamsType = paramsTypes[Config.FUNC_REQ_PARAMS_COUNT_LIMIT];
            for (int i = Config.FUNC_REQ_PARAMS_COUNT_LIMIT + 1; i < paramsCount; i++) {
                if (paramsTypes[i] != optParamsType) {
                    return null;
                }
            }
            actEncParamsTypes = EncodedTypeList.encodeTypeList(
                paramsTypes, Config.FUNC_REQ_PARAMS_COUNT_LIMIT, optParamsType);
        }
        return lookupFunctionForEncoded(actEncParamsTypes);
    }

    /**
     * Searches the group for a function that can accept parameters with specified types. The
     * actual parameters types are specified using encoded form.
     *
     * @param actEncParamsTypes encoded types of the actual parameters to be supported
     * @return
     *     a function from the group which supports specified actual parameters or
     *     <code>null</code> if no such function exists
     */
    protected abstract Function lookupFunctionForEncoded(long actEncParamsTypes);

    /**
     * Checks whether functions in the given array may form a group of functions overloading
     * single name. Group of functions must obey the following rules:
     * <ul>
     * <li>must not be empty</li>
     * <li>all functions must have the same name</li>
     * <li>types of functions' parameters must be disjunctive</li>
     * </ul>
     * @param funcs array of functions that belong to single group
     * @throws IllegalArgumentException
     *     in case functions in the array doesn't obey rules for function's group: it must not
     *     be empty, all functions within group must have the same name, all functions must
     *     have disjunctive parameters' types
     * @see #areDisjunctive(FunctionSignature, FunctionSignature)
     */
    private static void checkGroup(Function[] funcs) {
        int funcsCount = funcs.length;
        for (int i = 0; i < funcsCount; i++) {
            for (int j = i + 1; j < funcsCount; j++) {
                if (!areDisjunctive(funcs[i].getSignature(), funcs[j].getSignature())) {
                    throw new IllegalArgumentException(
                        "Functions within single group must have disjunctive parameters' types"
                        + " (there's a collision between "
                        + funcs[i].getName()
                        + funcs[i].getSignature().getParametersTypesDescription()
                        + " and " + funcs[j].getName()
                        + funcs[j].getSignature().getParametersTypesDescription()
                        + ")");
                }
            }
        }
    }

    /**
     * Checks whether two given signatures are disjunctive, i.e. such that there's no sequence
     * of types of actual values that matches both of them.
     * <p>
     * For example, given two signatures:
     * <ul>
     * <li>
     * signature with required parameters of types: {@link Type#INTEGER}, {@link Type#BOOLEAN}
     * and optional parameters of type {@link Type#STRING}
     * </li>
     * <li>
     * signature with required parameters of types: {@link Type#INTEGER}, {@link Type#BOOLEAN},
     * {@link Type#STRING}, {@link Type#STRING} and without optional parameters
     * </li>
     * </ul>
     * there's a sequence of types of actual parameters that matches both of them, e.g.:
     * {@link Type#INTEGER}, {@link Type#BOOLEAN}, {@link Type#STRING} so these signatures are
     * not disjunctive.
     *
     * @param aSignature first signature
     * @param bSignature second signature
     * @return
     *     <code>true</code> if given signatures are disjunctive, <code>false</code> otherwise
     */
    private static boolean areDisjunctive(
        FunctionSignature aSignature, FunctionSignature bSignature) {
        Type[] aReqParamsTypes = aSignature.getRequiredParametersTypes();
        Type[] bReqParamsTypes = bSignature.getRequiredParametersTypes();
        Type aOptParamsType = aSignature.getOptionalParametersType();
        Type bOptParamsType = bSignature.getOptionalParametersType();
        boolean differenceFound = false;
        int minReqParams = Math.min(aReqParamsTypes.length, bReqParamsTypes.length);
        for (int i = 0; i < minReqParams; i++) {
            if (aReqParamsTypes[i] != bReqParamsTypes[i]) {
                differenceFound = true;
                break;
            }
        }
        for (int i = aReqParamsTypes.length; i < bReqParamsTypes.length; i++) {
            if (bReqParamsTypes[i] != aOptParamsType) {
                differenceFound = true;
                break;
            }
        }
        for (int i = bReqParamsTypes.length; i < aReqParamsTypes.length; i++) {
            if (aReqParamsTypes[i] != bOptParamsType) {
                differenceFound = true;
                break;
            }
        }
        return differenceFound;
    }


    /**
     * Specialization of {@link FunctionsGroup} that contains only one function. By using it
     * we can minimize overhead caused by overloaded functions' names.
     *
     * @author Zbigniew Chyla (zbigniew@chyla.pl)
     */
    private static final class SingletonGroup extends FunctionsGroup {

        /**
         * Function that belongs to the group
         */
        private final Function func;

        /**
         * Encoded declared types of parameters of {@link #func}
         */
        private final long declEncParamsTypes;

        /**
         * Constructs a group containing given function
         *
         * @param func a function to place in the group
         */
        SingletonGroup(Function func) {
            this.func = func;
            FunctionSignature signature = func.getSignature();
            this.declEncParamsTypes = EncodedTypeList.encodeTypeList(
                signature.getRequiredParametersTypes(),
                signature.getRequiredParametersTypes().length,
                signature.getOptionalParametersType());
        }

        /**
         * Returns a name of the group using a name of the only contained function
         *
         * @return name of the group
         */
        String getName() {
            return func.getName();
        }

        /**
         * {@inheritDoc}
         *
         * @param actEncParamsTypes {@inheritDoc}
         */
        protected Function lookupFunctionForEncoded(long actEncParamsTypes) {
            if (EncodedTypeList.isTypeListMatching(declEncParamsTypes, actEncParamsTypes)) {
                return func;
            } else {
                return null;
            }
        }

    }


    /**
     * Specialisation of {@link FunctionsGroup} that may contain any number of functions.
     *
     * @author Zbigniew Chyla (zbigniew@chyla.pl)
     */
    private static final class FullGroup extends FunctionsGroup {

        /**
         * Functions that belong to the group
         */
        private final Function[] funcs;

        /**
         * Encoded declared types of parameters of functions that belong to the group, in order
         * identical to {@link #funcs}
         */
        private final long[] declEncParamsTypes;

        /**
         * Constructs a group containing given functions
         *
         * @param funcs functions to place in the group
         */
        FullGroup(Function[] funcs) {
            this.funcs = funcs;
            this.declEncParamsTypes = new long[funcs.length];
            for (int i = 0; i < funcs.length; i++) {
                FunctionSignature signature = funcs[i].getSignature();
                this.declEncParamsTypes[i] = EncodedTypeList.encodeTypeList(
                    signature.getRequiredParametersTypes(),
                    signature.getRequiredParametersTypes().length,
                    signature.getOptionalParametersType());
            }
        }

        /**
         * Returns a name of the group using a name of the first contained function
         *
         * @return name of the group
         */
        String getName() {
            return funcs[0].getName();
        }

        /**
         * {@inheritDoc}
         *
         * @param actEncParamsTypes {@inheritDoc}
         */
        protected Function lookupFunctionForEncoded(long actEncParamsTypes) {
            int funcIdx = EncodedTypeList.findMatchingTypeList(
                declEncParamsTypes, actEncParamsTypes);
            if (funcIdx >= 0) {
                return funcs[funcIdx];
            } else {
                return null;
            }
        }

    }

}
