/*
 * Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License version 2.1 as published by the Free Software
 * Foundation.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301  USA
 */
package pl.chyla.wyjec.stdconsts;

import java.util.Calendar;

import pl.chyla.wyjec.core.IConstantsParser;
import pl.chyla.wyjec.core.Value;
import pl.chyla.wyjec.utils.CharUtils;
import pl.chyla.wyjec.utils.ReadableRawString;
import pl.chyla.wyjec.utils.SimpleCalendar;


/**
 * Parser for date constants. Date constant is represented with "@" character followed by date
 * specification in format specified by ISO 8601 standard with optional time of the day part.
 * Time of day part (if present) must be separated from date part with single space. In case
 * time of the day part is not present, midnight (00:00:00) is assumed.
 * <p>
 * In short, constant format must be one of:
 * <ul>
 * <li><code>@YYYY-MM-DD</code></li>
 * <li><code>@YYYY-MM-DD hh:mm:ss</code></li>
 * </ul>
 *
 * @see IConstantsParser
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class DateConstantsParser implements IConstantsParser {

    /**
     * First character of constant representing date value.
     */
    private static final char DATE_PREFIX = '@';

    /**
     * Position of year number in date constant, relative to position the first character.
     */
    private static final int YEAR_POS = 1;

    /**
     * Number of characters occupied by YEAR field
     */
    private static final int YEAR_CHARS_COUNT = 4;

    /**
     * Character following year number.
     */
    private static final char YEAR_TERMINATOR = '-';

    /**
     * Position of character following year number in date constant, relative to position the
     * first character.
     */
    private static final int YEAR_TERMINATOR_POS = YEAR_POS + YEAR_CHARS_COUNT;

    /**
     * Position of month number in date constant, relative to position the first character.
     */
    private static final int MONTH_POS = YEAR_TERMINATOR_POS + 1;

    /**
     * Number of characters occupied by MONTH field
     */
    private static final int MONTH_CHARS_COUNT = 2;

    /**
     * Character following month number.
     */
    private static final char MONTH_TERMINATOR = '-';

    /**
     * Position of character following month number in date constant, relative to position the
     * first character.
     */
    private static final int MONTH_TERMINATOR_POS = MONTH_POS + MONTH_CHARS_COUNT;

    /**
     * Position of day of month number in date constant, relative to position the first
     * character.
     */
    private static final int DAY_POS = MONTH_TERMINATOR_POS + 1;

    /**
     * Number of characters occupied by DAY field
     */
    private static final int DAY_CHARS_COUNT = 2;

    /**
     * Character following day number in case constant is represented using long format.
     */
    private static final char DAY_TERMINATOR = ' ';

    /**
     * Position of character following day number in date constant in case constant is
     * represented using long format, relative to position the  first character.
     */
    private static final int DAY_TERMINATOR_POS = DAY_POS + DAY_CHARS_COUNT;

    /**
     * Position of hour number in date constant, relative to position the first character.
     */
    private static final int HOUR_POS = DAY_TERMINATOR_POS + 1;

    /**
     * Number of characters occupied by HOUR field
     */
    private static final int HOUR_CHARS_COUNT = 2;

    /**
     * Character following hour number.
     */
    private static final char HOUR_TERMINATOR = ':';

    /**
     * Position of character following hour number in date constant, relative to position the
     * first character.
     */
    private static final int HOUR_TERMINATOR_POS = HOUR_POS + HOUR_CHARS_COUNT;

    /**
     * Position of minute number in date constant, relative to position the first character.
     */
    private static final int MINUTE_POS = HOUR_TERMINATOR_POS + 1;

    /**
     * Number of characters occupied by MINUTE field
     */
    private static final int MINUTE_CHARS_COUNT = 2;

    /**
     * Character following minute number.
     */
    private static final char MINUTE_TERMINATOR = ':';

    /**
     * Position of character following minute number in date constant, relative to position the
     * first character.
     */
    private static final int MINUTE_TERMINATOR_POS = MINUTE_POS + MINUTE_CHARS_COUNT;

    /**
     * Position of second number in date constant, relative to position the first character.
     */
    private static final int SECOND_POS = MINUTE_TERMINATOR_POS + 1;

    /**
     * Number of characters occupied by SECOND field
     */
    private static final int SECOND_CHARS_COUNT = 2;

    /**
     * Length of date constant using short format, i.e. <code>@YYYY-MM-DD</code>
     */
    private static final int SHORT_DATE_LENTGH = DAY_TERMINATOR_POS;

    /**
     * Length of date constant using long format, i.e. <code>@YYYY-MM-DD hh:mm:ss</code>
     */
    private static final int LONG_DATE_LENTGH = SECOND_POS + SECOND_CHARS_COUNT;

    /**
     * Temporary calendar used for calculations
     */
    private final SimpleCalendar tmpCalendar = new SimpleCalendar();

    /**
     * Private (empty) constructor to disallow instantiating of the class from outside. Class'
     * instance should be accessed via {@link #getInstance()} getter.
     */
    private DateConstantsParser() { }

    /**
     * Returns instance of <code>DateConstantsParser</code> which can be safely used independly
     * from other threads which also get <code>DateConstantsParser</code> using this method.
     * It's undefined whether this method returns newly created instance or not.
     *
     * @return instance of <code>DateConstantsParser</code>
     */
    public static DateConstantsParser getInstance() {
        return new DateConstantsParser();
    }

    /**
     * Parses given string as a representation of date constant value. If the name contains
     * string representation of date, using <code>@YYYY-MM-DD</code> or
     * <code>@YYYY-MM-DD hh:mm:ss</code> format, its value will be stored in given {@link Value}
     * and can be later accessed using {@link Value#getDate()} method.
     *
     * @param result {@link Value} object used to store recognized date value
     * @param constStr string representation of constant to parse
     * @return
     *     <code>true</code> if given string contains textual representation of date value (in
     *     this case its value can be accessed using given {@link Value} object),
     *     <code>false</code> otherwise.
     * @see IConstantsParser#parseConstant(Value, ReadableRawString)
     */
    public boolean parseConstant(Value result, ReadableRawString constStr) {
        char[] chars = constStr.getStorage();
        int startPos = constStr.getStartPos();
        int nameLen = constStr.getLength();
        if (!isValidDate(chars, startPos, nameLen)) {
            return false;
        }
        int year = parseNDigitsNumber(chars, startPos + YEAR_POS, YEAR_CHARS_COUNT);
        int month = parseNDigitsNumber(chars, startPos + MONTH_POS, MONTH_CHARS_COUNT) - 1;
        int day = parseNDigitsNumber(chars, startPos + DAY_POS, DAY_CHARS_COUNT);
        if (nameLen == LONG_DATE_LENTGH) {
            int hour = parseNDigitsNumber(chars, startPos + HOUR_POS, HOUR_CHARS_COUNT);
            int minute = parseNDigitsNumber(chars, startPos + MINUTE_POS, MINUTE_CHARS_COUNT);
            int second = parseNDigitsNumber(chars, startPos + SECOND_POS, SECOND_CHARS_COUNT);
            tmpCalendar.setDate(year, month, day, hour, minute, second);
        } else {
            tmpCalendar.setDate(year, month, day, 0, 0, 0);
        }
        tmpCalendar.set(Calendar.MILLISECOND, 0);
        result.setDate(tmpCalendar.getTimeInMillis());
        return true;
    }

    /**
     * Checks whether given sequence of characters (specified as part of given array) contains
     * valid date constant of form <code>@YYYY-MM-DD hh:mm:ss</code> or
     * <code>@YYYY-MM-DD hh:mm:ss</code>.
     *
     * @param chars array containing sequence of characters to check
     * @param startPos position of first character to check in the array
     * @param nameLen number of characters forming the sequence to check
     * @return
     *     <code>true</code> if specified sequence of characters contains valid date constant,
     *     <code>false</code> otherwise.
     */
    private boolean isValidDate(char[] chars, int startPos, int nameLen) {
        if ((nameLen != SHORT_DATE_LENTGH && nameLen != LONG_DATE_LENTGH)
            || chars[startPos] != DATE_PREFIX) {
            return false;
        }
        boolean isValid =
            chars[startPos + YEAR_TERMINATOR_POS] == YEAR_TERMINATOR
            && chars[startPos + MONTH_TERMINATOR_POS] == MONTH_TERMINATOR;
        isValid &=
            isValidNDigitsNumber(chars, startPos + YEAR_POS, YEAR_CHARS_COUNT)
            && isValidNDigitsNumber(chars, startPos + MONTH_POS, MONTH_CHARS_COUNT)
            && isValidNDigitsNumber(chars, startPos + DAY_POS, DAY_CHARS_COUNT);
        if (nameLen == LONG_DATE_LENTGH) {
            isValid &=
                chars[startPos + DAY_TERMINATOR_POS] == DAY_TERMINATOR
                && chars[startPos + HOUR_TERMINATOR_POS] == HOUR_TERMINATOR
                && chars[startPos + MINUTE_TERMINATOR_POS] == MINUTE_TERMINATOR;
            isValid &=
                isValidNDigitsNumber(chars, startPos + HOUR_POS, HOUR_CHARS_COUNT)
                && isValidNDigitsNumber(chars, startPos + MINUTE_POS, MINUTE_CHARS_COUNT)
                && isValidNDigitsNumber(chars, startPos + SECOND_POS, SECOND_CHARS_COUNT);
        }
        return isValid;
    }

    /**
     * Checks whether given n-element sequence of characters (specified as part of given array)
     * contains a valid n-digit decimal number.
     *
     * @param chars array containing sequence of characters to check
     * @param startPos position of first character to check in the array
     * @param digitsCount number of digits to check
     * @return
     *     <code>true</code> if specified sequence of characters contains valid n-digit number,
     *     <code>false</code> otherwise.
     */
    private boolean isValidNDigitsNumber(char[] chars, int startPos, int digitsCount) {
        boolean isValid = true;
        for (int i = 0; i < digitsCount; i++) {
            if (!CharUtils.isDigit(chars[startPos + i])) {
                isValid = false;
                break;
            }
        }
        return isValid;
    }

    /**
     * Parses n-character part of given array as n-digit string representation of integer value.
     *
     * @param chars array containing sequence of characters to parse
     * @param startPos position of first character to parse in the array
     * @param digitsCount number of digits to parse
     * @return (nonnegative) value of parsed number
     */
    private int parseNDigitsNumber(char[] chars, int startPos, int digitsCount) {
        int result = 0;
        for (int i = 0; i < digitsCount; i++) {
            result = result * 10 + chars[startPos + i] - '0';
        }
        return result;
    }

}
