/*
 * Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License version 2.1 as published by the Free Software
 * Foundation.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301  USA
 */
package pl.chyla.wyjec.stdfuncs.datefuncs.basic;

import pl.chyla.wyjec.core.FunctionSignature;
import pl.chyla.wyjec.core.Parameter;
import pl.chyla.wyjec.core.functions.DateFunction;
import pl.chyla.wyjec.core.parameters.IntegerParameter;
import pl.chyla.wyjec.utils.SimpleCalendar;


/**
 * Returns date value that corresponds to given year, month, day of the month and time of
 * the day specified with hour, minute and second.
 *
 * @wyjec.func date
 * @wyjec.example "{date|{1978}|{09}|{11}|{17}|{18}|{3}}" "1978-09-11 17:18:03" date
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class FnDateFromYMDHMS extends DateFunction {

    /**
     * Year number of the date to use
     * @wyjec.param year
     */
    private static final IntegerParameter pYear = new IntegerParameter();

    /**
     * Month number of the date to use
     * @wyjec.param month
     */
    private static final IntegerParameter pMonth = new IntegerParameter();

    /**
     * Day of the month of the date to use
     * @wyjec.param day
     */
    private static final IntegerParameter pDay = new IntegerParameter();

    /**
     * Hour part in time of the day of the date to use
     * @wyjec.param hour
     */
    private static final IntegerParameter pHour = new IntegerParameter();

    /**
     * Minute part in time of the day of the date to use
     * @wyjec.param minute
     */
    private static final IntegerParameter pMinute = new IntegerParameter();

    /**
     * Second part in time of the day of the date to use
     * @wyjec.param minute
     */
    private static final IntegerParameter pSecond = new IntegerParameter();

    private static final FunctionSignature signature = new FunctionSignature(
        new Parameter[] {pYear, pMonth, pDay, pHour, pMinute, pSecond}, null);

    /**
     * Temporary calendar instance used for calculation
     */
    private SimpleCalendar tmpCal = new SimpleCalendar();

    public FunctionSignature getSignature() {
        return signature;
    }

    public String getName() {
        return "date";
    }

    protected long execute() {
        int year = (int) getParameter(pYear);
        int month = (int) getParameter(pMonth) - 1;
        int day = (int) getParameter(pDay);
        int hour = (int) getParameter(pHour);
        int minute = (int) getParameter(pMinute);
        int second = (int) getParameter(pSecond);
        tmpCal.setDate(year, month, day, hour, minute, second);
        return tmpCal.getTimeInMillis();
    }

}
