/*
 * Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License version 2.1 as published by the Free Software
 * Foundation.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301  USA
 */
package pl.chyla.wyjec.stdfuncs.stringfuncs.ops;

import pl.chyla.wyjec.core.FunctionSignature;
import pl.chyla.wyjec.core.Parameter;
import pl.chyla.wyjec.core.functions.StringFunction;
import pl.chyla.wyjec.core.parameters.StringParameter;
import pl.chyla.wyjec.utils.RawString;
import pl.chyla.wyjec.utils.ReadableRawString;


/**
 * Returns a copy of given string value with specified characters stripped from its suffix.
 *
 * @wyjec.func rstrip
 * @wyjec.example "{rstrip|__.abc.__|._}" "__.abc" string
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class FnRStrip extends StringFunction {

    /**
     * Base string used to generate the new one
     * @wyjec.param value
     */
    private static final StringParameter pValue = new StringParameter();

    /**
     * String containing characters that are to be stripped from the base string's suffix
     * @wyjec.param stripChars
     */
    private static final StringParameter pStripChars = new StringParameter();

    private static final FunctionSignature signature = new FunctionSignature(
        new Parameter[] {pValue, pStripChars}, null);

    public FunctionSignature getSignature() {
        return signature;
    }

    public String getName() {
        return "rstrip";
    }

    /**
     * Counts characters to strip from the end of given string value.
     *
     * @param value string value which suffix is to be removed
     * @param stripChars string containing characters that are to be stripped from the suffix
     * @return number of characters to be stripped
     */
    private int countStrippedChars(ReadableRawString value, ReadableRawString stripChars) {
        char[] chars = value.getStorage();
        int startPos = value.getStartPos();
        int i = value.getEndPos() - 1;
        while (i >= startPos && stripChars.find(chars[i], 0) >= 0) {
            i--;
        }
        return value.getEndPos() - (i + 1);
    }

    protected void execute(RawString result) {
        ReadableRawString value = getParameter(pValue);
        int strippedCharsCount = countStrippedChars(value, getParameter(pStripChars));
        result.append(value, 0, value.getLength() - strippedCharsCount);
    }

}
