/*
 * Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License version 2.1 as published by the Free Software
 * Foundation.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301  USA
 */
package pl.chyla.wyjec.stdfuncs.stringfuncs.ops;

import pl.chyla.wyjec.core.FunctionSignature;
import pl.chyla.wyjec.core.Parameter;
import pl.chyla.wyjec.core.functions.StringFunction;
import pl.chyla.wyjec.core.parameters.IntegerParameter;
import pl.chyla.wyjec.core.parameters.StringParameter;
import pl.chyla.wyjec.utils.RawString;
import pl.chyla.wyjec.utils.ReadableRawString;


/**
 * Returns substring of given string. Substring is specified with 0-based index of its first
 * character and its length, where negative indices allow for counting characters from the end
 * of the string). Invalid indices will be bound to the nearest end of the string. Negative
 * length will be treated as zero.
 *
 * @wyjec.func substr
 * @wyjec.example "{substr|Kowalski|{4}|{2}}" "ls" string
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class FnSubStr extends StringFunction {

    /**
     * String to generate a substring from
     * @wyjec.param value
     */
    private static final StringParameter pValue = new StringParameter();

    /**
     * Index of first character to include in the substring
     * @wyjec.param subStart
     */
    private static final IntegerParameter pSubStart = new IntegerParameter();

    /**
     * Length (number of characters) of the substring
     * @wyjec.param subLength
     */
    private static final IntegerParameter pSubLength = new IntegerParameter();

    private static final FunctionSignature signature = new FunctionSignature(
        new Parameter[] {pValue, pSubStart, pSubLength}, null);

    public FunctionSignature getSignature() {
        return signature;
    }

    public String getName() {
        return "substr";
    }

    protected void execute(RawString result) {
        ReadableRawString str = getParameter(pValue);
        int strLen = str.getLength();
        int subStart = (int) getParameter(pSubStart);
        if (subStart < 0) {
            subStart = Math.max(strLen + subStart, 0);
        }
        subStart = Math.min(subStart, strLen);
        int subLen = (int) getParameter(pSubLength);
        if (subLen > 0) {
            subLen = Math.min(subLen, strLen - subStart);
            result.append(str, subStart, subLen);
        }
    }

}
