/*
 * Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License version 2.1 as published by the Free Software
 * Foundation.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301  USA
 */
package pl.chyla.wyjec.utils;

import java.util.ArrayList;


/**
 * Helper class grouping simple operations on arrays.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class ArrayUtils {

    /**
     * Private empty constructor to disallow instantiating of the class.
     */
    private ArrayUtils() { }

    /**
     * Copies all elements from one array of objects to another storing them at indices
     * starting from 0. Destination array must not be smaller than the source one and it must
     * be assign-compatible with source.
     *
     * @param src array to copy elements from
     * @param dest array to copy elements to
     */
    public static void copyArray(Object[] src, Object[] dest) {
        System.arraycopy(src, 0, dest, 0, src.length);
    }

    /**
     * Returns new array of characters containing only characters from specified part of
     * another array.
     *
     * @param src base array of characters to copy characters from
     * @param srcStart
     *     index of first character in <code>src</code> array to copy into new array
     * @param srcLength number of characters to copy from <code>src</code> into new array
     * @return
     *     new array of characters of length <code>srcLength</code>, containing specified
     *     characters from <code>src</code>
     */
    public static char[] getCharSubArray(char[] src, int srcStart, int srcLength) {
        char[] dest = new char[srcLength];
        System.arraycopy(src, srcStart, dest, 0, srcLength);
        return dest;
    }

    /**
     * Flattens given array of arrays by creating the new array containing elements from source
     * sub-arrays.
     *
     * @param srcArrays
     *     source array of arrays; both top-level array and sub-arrays may be empty, but neither
     *     of them may be equal to <code>null</code>; runtime types of all sub-arrays must be
     *     assign-compatible with <code>destSample</code>
     * @param destSample
     *     an array specifying runtime type of an array returned as the result; its contents
     *     will be ignored so it's recommended to pass just an empty array
     * @return
     *     new array containing references to all objects from direct sub-arrays of given source
     *     array, with the same runtime type as <code>destSample</code> parameter
     */
    public static Object[] flattenArrays(Object[][] srcArrays, Object[] destSample) {
        int elementsCount = 0;
        for (int i = 0; i < srcArrays.length; i++) {
            elementsCount += srcArrays[i].length;
        }
        ArrayList elements = new ArrayList(elementsCount);
        for (int i = 0; i < srcArrays.length; i++) {
            Object[] srcArray = srcArrays[i];
            for (int j = 0; j < srcArray.length; j++) {
                elements.add(srcArray[j]);
            }
        }
        return elements.toArray(destSample);
    }

}
