/*
 * Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License version 2.1 as published by the Free Software
 * Foundation.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301  USA
 */
package pl.chyla.wyjec.utils;


/**
 * Utility class allowing for fast writing of string representation of <code>int</code> number
 * into existing array of characters.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class IntFormatter {

    /**
     * Maximum number of digits for string representation of nonnegative number
     * of type <code>int</code>
     */
    public static final int MAX_DIGITS_COUNT = 10;

    /**
     * Maximum number of characters (including sign character) for string representation
     * of number of type <code>int</code>
     */
    public static final int MAX_CHARACTERS_COUNT = MAX_DIGITS_COUNT + 1;

    /**
     * Array of characters that form string representation of Integer.MIN_VALUE value
     */
    private static final char[] MIN_INT_CHARACTERS =
        Integer.toString(Integer.MIN_VALUE).toCharArray();

    /**
     * Private empty constructor to disallow instantiating of the class.
     */
    private IntFormatter() { }

    /**
     * Stores decimal string representation of given number in consecutive components of
     * specified array of characters. Characters will be written in the array starting from
     * specified index.
     *
     * @param value value to generate string representation for
     * @param dest array to write string representation to
     * @param startPos
     *     index of first component of <code>dest</code> array to use for writing characters
     * @return
     *     number of characters written to the array or 0 if string representation doesn't fit
     *     in specified part of the array (no character is written in this case)
     */
    public static int formatToArray(int value, char[] dest, int startPos) {
        int freeChars = dest.length - startPos;
        int storedCharsCount;
        if (value >= 0) {
            int charsCount = getNonnegativeDigitsCount(value);
            if (charsCount <= freeChars) {
                storeDigits(value, dest, startPos, charsCount);
                storedCharsCount = charsCount;
            } else {
                storedCharsCount = 0;
            }
        } else if (value != Integer.MIN_VALUE) {
            int posValue = -value;
            int posCharsCount = getNonnegativeDigitsCount(posValue);
            if (posCharsCount < freeChars) {
                dest[startPos] = '-';
                storeDigits(posValue, dest, startPos + 1, posCharsCount);
                storedCharsCount = posCharsCount + 1;
            } else {
                storedCharsCount = 0;
            }
        } else {
            if (MIN_INT_CHARACTERS.length <= freeChars) {
                System.arraycopy(
                    MIN_INT_CHARACTERS, 0, dest, startPos, MIN_INT_CHARACTERS.length);
                storedCharsCount = MIN_INT_CHARACTERS.length;
            } else {
                storedCharsCount = 0;
            }
        }
        return storedCharsCount;
    }

    /**
     * Helper method: stores digits of given nonnegative value into given array starting from
     * specified position.
     *
     * @param value value to generate digits from
     * @param dest array to write digits to
     * @param startPos index of first digit in <code>dest</code> array
     * @param charsCount number of digits to store
     */
    private static void storeDigits(int value, char[] dest, int startPos, int charsCount) {
        int v = value;
        for (int i = startPos + charsCount - 1; i >= startPos; i--) {
            dest[i] = (char) ('0' + v % 10);
            v /= 10;
        }
    }

    /**
     * Measures length of decimal string representation of given number without actually
     * generating it.
     *
     * @param value value to analyse
     * @return
     *     length of string representation of given number (in characters, including sign
     *     character in case of negative numbers)
     */
    public static int getCharactersCount(int value) {
        int count;
        if (value >= 0) {
            count = getNonnegativeDigitsCount(value);
        } else if (value != Integer.MIN_VALUE) {
            count = getCharactersCount(-value) + 1;
        } else {
            count = MAX_CHARACTERS_COUNT;
        }
        return count;
    }

    /**
     * Measures length of decimal string representation of given nonnegative number without
     * actually generating it.
     *
     * @param value value to analyse
     * @return
     *     length of string representation of given number (in characters, only decimal digits)
     */
    private static int getNonnegativeDigitsCount(int value) {
        int digitsCount = 1;
        int upperLimit = 10;
        while (value >= upperLimit && digitsCount < MAX_DIGITS_COUNT) {
            digitsCount++;
            upperLimit *= 10;
        }
        return digitsCount;
    }

}
