/*
 * Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License version 2.1 as published by the Free Software
 * Foundation.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301  USA
 */
package pl.chyla.wyjec.utils;

import java.util.EmptyStackException;


/**
 * Stack (LIFO structure) of <code>int</code> values.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class IntStack {

    /**
     * Default size of <code>elements</code> array
     */
    private static final int DEFAULT_SIZE = 16;

    /**
     * Array containing elements on the stack
     */
    private int[] elements;

    /**
     * Current size of the stack (number of elements used in <code>elements</code> array)
     */
    private int size;

    /**
     * Constructs empty stack of <code>int</code> values
     */
    public IntStack() {
        elements = new int[DEFAULT_SIZE];
        size = 0;
    }

    /**
     * Constructs empty stack of <code>int</code> values with pre-allocated space for given
     * number of elements. An instance created using this constructor give a warranty that no
     * extra memory allocations will be performed until total number of elements on the stack
     * exceeds given number.
     *
     * @param initialCapacity initial capacity of this stack
     * @throws IllegalArgumentException in case given initial capacity is negative
     */
    public IntStack(int initialCapacity) {
        if (initialCapacity < 0) {
            throw new IllegalArgumentException(
                "Initial stack capacity must not be negative: " + initialCapacity);
        }
        elements = new int[initialCapacity];
        size = 0;
    }

    /**
     * Assures given minimum capacity of array containing elements on the stack (by reallocating
     * it if needed).
     *
     * @param minCapacity requested minimum capacity of the stack
     */
    private void ensureCapacity(int minCapacity) {
        if (elements.length < minCapacity) {
            int newCapacity = Math.max(elements.length * 2, minCapacity);
            int[] newElements = new int[newCapacity];
            System.arraycopy(elements, 0, newElements, 0, elements.length);
            elements = newElements;
        }
    }

    /**
     * Pushes given number on the top of this stack.
     *
     * @param value value to push on the stack
     */
    public void push(int value) {
        size++;
        ensureCapacity(size);
        elements[size - 1] = value;
    }

    /**
     * Removes number at the top of this stack and returns its value.
     *
     * @return value removed from the top of the stack
     * @throws EmptyStackException in case the stack is empty
     */
    public int pop() {
        if (size == 0) {
            throw new EmptyStackException();
        }
        size--;
        return elements[size];
    }

    /**
     * Returns value of number at the top of this stack.
     *
     * @return value at the top of this stack
     * @throws EmptyStackException in case the stack is empty
     */
    public int peek() {
        if (size == 0) {
            throw new EmptyStackException();
        }
        return elements[size - 1];
    }

    /**
     * Informs whether this stack is empty.
     *
     * @return <code>true</code> if the stack is empty, <code>false</code> otherwise
     */
    public boolean isEmpty() {
        return size == 0;
    }

}
