/*
 * Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License version 2.1 as published by the Free Software
 * Foundation.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301  USA
 */
package pl.chyla.wyjec.utils;

import java.util.Stack;


/**
 * Objects pool keeping constant number of preallocated objects. In case the number of objects
 * requested from this pool exceeds its size, new objects will be created dynamically using a
 * factory passed to the pool during its construction. Unless its size is exceeded, this pool
 * guarantees that no object constructions/destructions will happen and executing
 * {@link #getObject()} and {@link #releaseObject(Object)} operations will consume constant time
 * (independent of the size of the pool, number of requested objects, etc.).
 * <p>
 * <code>ObjectsPool</code> instances may be used as an alternative for dynamic allocation of
 * objects using <code>new</code> statement. It provides a method for getting "new" instance of
 * supported type and another method for returning an instance when it's no longer needed.
 * The design of this interface is based on assumption that program using it typically performs
 * its processing in "sessions", i.e. after executing some group of operations all objects
 * acquired from the pool may be freed (they're no longer needed).
 * <p>
 * Typical fragment of program using <code>ObjectsPool</code> follows:
 * <pre>
 * ObjectsPool pool = new ObjectsPool(..., ...);
 * // ...
 * try {
 *     // Execute some operations, getting objects with {@link #getObject() pool.getObject()}
 *     // and then releasing them with {@link #releaseObject(Object) pool.releaseObject(o)}.
 * } finally {
 *     pool.resetPool()
 * }
 * </pre>
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class ObjectsPool {

    /**
     * Factory used to generate new objects when there are no available objects in this pool.
     */
    private final IObjectsFactory factory;

    /**
     * All objects that belong to the pool, they are created during initialization and never
     * destroyed while the pool is alive.
     */
    private final Object[] ownObjects;

    /**
     * Array of available objects used as primary place for storing references to currently
     * available objects. Only components with indices less than {@link #availableStaticCount}
     * are actually used. On initialization this array gets filled with references copied from
     * {@link #ownObjects} but during operation it may contain both references from
     * <code>ownObjects</code> and references to dynamically created objects.
     */
    private final Object[] availableStatic;

    /**
     * Number of currently available elements in {@link #availableStatic} array.
     */
    private int availableStaticCount;

    /**
     * Number of elements from the beginning of {@link #availableStatic} array that are
     * identical to their counterparts in <code>ownObjects</code>. Thanks to keeping and
     * updating this value, on resetting only elements with greater indices have to be copied
     * from <code>ownObjects</code>.
     */
    private int ownStaticCount;

    /**
     * Stack used for storing references to currently available objects that don't fit into
     * {@link #availableStatic} array. This stack will be used only if the number of objects
     * acquired from this pool exceeds the size of {@link #availableStatic} array.
     */
    private final Stack availableDynamic;

    /**
     * Constructs new pool with specified number of statically allocated objects. These objects
     * will be instantiated by the constructor and will not be destroyed before destruction of
     * this <code>ObjectsPool</code> instance.
     *
     * @param factory
     *     factory used to create objects (both created statically at initialization
     *     phase and dynamically on demand)
     * @param size number of statically created objects
     * @throws IllegalArgumentException in case given size is negative
     */
    public ObjectsPool(IObjectsFactory factory, int size) {
        if (size < 0) {
            throw new IllegalArgumentException(
                "Initial pool size must not be negative: " + size);
        }
        this.factory = factory;
        ownObjects = new Object[size];
        for (int i = 0; i < size; i++) {
            ownObjects[i] = factory.createObject();
        }
        availableStatic = new Object[size];
        ArrayUtils.copyArray(ownObjects, availableStatic);
        availableStaticCount = availableStatic.length;
        ownStaticCount = availableStaticCount;
        availableDynamic = new Stack();
    }

    /**
     * Fetches new object from the pool. It should be later returned to the pool with
     * {@link #releaseObject(Object)} method. Alternatively, all fetched objects will be
     * returned at once with {@link #reset()} method.
     *
     * @return object acquired from this pool
     */
    public Object getObject() {
        Object obj;
        if (availableStaticCount > 0) {
            availableStaticCount--;
            obj = availableStatic[availableStaticCount];
        } else if (!availableDynamic.isEmpty()) {
            obj = availableDynamic.pop();
        } else {
            obj = factory.createObject();
        }
        return obj;
    }

    /**
     * Returns given object to this pool. The object should be one of objects previously
     * acquired with {@link #getObject()} method from this pool. After calling this method,
     * using the object given as a parameter is forbidden. An object may be returned only to a
     * poll from which it was previously acquired, returning it to other pools (even if they
     * use the same factory of objects) is forbidden.
     *
     * @param obj object to return to this pool
     */
    public void releaseObject(Object obj) {
        if (availableStaticCount < availableStatic.length) {
            ownStaticCount = Math.min(ownStaticCount, availableStaticCount);
            availableStatic[availableStaticCount] = obj;
            availableStaticCount++;
        } else {
            availableDynamic.push(obj);
        }
    }

    /**
     * Resets this pool to its initial state. After calling this method, any objects previously
     * acquired from this pool may be used only until the next call to {@link #getObject()}.
     * In most cases it simply means that after calling this method using any previously
     * acquired objects is forbidden.
     */
    public void reset() {
        int dirtyCount = availableStatic.length - ownStaticCount;
        if (dirtyCount > 0) {
            System.arraycopy(
                ownObjects, ownStaticCount, availableStatic, ownStaticCount, dirtyCount);
            ownStaticCount = availableStatic.length;
        }
        availableStaticCount = availableStatic.length;
        availableDynamic.clear();
    }

}
