/*
 * Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License version 2.1 as published by the Free Software
 * Foundation.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301  USA
 */
package pl.chyla.wyjec.utils;

import java.util.Arrays;


/**
 * Concrete extension of ReadableRawString offering simple operations for editing string.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class RawString extends ReadableRawString {

    /**
     * Constructs new <code>RawString</code> instance with private storage of size equal to
     * length of given string and containing string equal to given string.
     * <p>
     * After construction the following will be true (<code>s</code> - constructed instance):
     * <ul>
     * <li><code>s.toString().equals(other)</code></li>
     * <li><code>s.getStorage().length == other.length()</code></li>
     * </ul>
     *
     * @param other string to use to initialize the instance
     */
    public RawString(String other) {
        set(other.toCharArray());
    }

    /**
     * Constructs new <code>RawString</code> instance with private storage of size 0 containing
     * empty string.
     * <p>
     * After construction the following will be true (<code>s</code> - constructed instance):
     * <ul>
     * <li><code>s.toString().equals("")</code></li>
     * <li><code>s.getStorage().length == 0</code></li>
     * </ul>
     */
    public RawString() {
        this(0);
    }

    /**
     * Constructs new <code>RawString</code> instance with private storage of given size
     * containing empty string (all characters in newly created storage will be unused).
     * <p>
     * After construction the following will be true (<code>s</code> - constructed instance):
     * <ul>
     * <li><code>s.toString().equals("")</code></li>
     * <li><code>s.getStorage().length == size</code></li>
     * <li><code>s.getStartPos() == 0</code></li>
     * </ul>
     *
     * @param size size of storage to create
     * @throws IllegalArgumentException in case <code>size</code> is negative
     */
    public RawString(int size) {
        if (size < 0) {
            throw new IllegalArgumentException(
                "Storage size must not be negative (" + size + " used)");
        }
        set(new char[size], 0, 0);
    }

    /**
     * Constructs new <code>RawString</code> instance with given array used as storage and
     * value formed from all characters from the array.
     * <p>
     * After construction the following will be true (<code>s</code> - constructed instance):
     * <ul>
     * <li><code>s.toString().equals(new String(characters))</code></li>
     * <li><code>s.getStorage() == characters</code></li>
     * </ul>
     *
     * @param characters characters to use as storage for the instance
     */
    public RawString(char[] characters) {
        set(characters);
    }

    /**
     * Constructs new <code>RawString</code> instance with given array used as storage and
     * value formed from specified range of characters from the array.
     * <p>
     * After construction the following will be true (<code>s</code> - constructed instance):
     * <ul>
     * <li><code>s.toString().equals(new String(characters, startPos, length))</code></li>
     * <li><code>s.getStorage() == characters</code></li>
     * <li><code>s.getStartPos() == startPos</code></li>
     * </ul>
     *
     * @param characters characters to use as storage for the instance
     * @param startPos
     *     index of first character in <code>characters</code> to use as this string's value
     * @param length
     *     number of characters from <code>characters</code> to use as this string's value
     * @throws IllegalArgumentException
     *     in case <code>startPos</code> or <code>length</code> is negative or their sum is
     *     grater than the length of <code>characters</code> array
     */
    public RawString(char[] characters, int startPos, int length) {
        set(characters, startPos, length);
    }

    /**
     * Constructs new <code>RawString</code> instance with storage shared with given another
     * instance of <code>ReadableRawString</code> and equal value.
     * <p>
     * After construction the following will be true (<code>s</code> - constructed instance):
     * <ul>
     * <li><code>s.equals(other)</code></li>
     * <li><code>s.getStorage() == other.getStorage()</code></li>
     * </ul>
     *
     * @param other string to use to initialize the instance
     */
    public RawString(ReadableRawString other) {
        set(other);
    }

    /**
     * Sets storage of this string to given array of characters and forms new value from all
     * characters of the array.
     * <p>
     * After calling this method the instance will be in the same state, as if it were created
     * with <code>new RawString(newStorage)</code>.
     *
     * @param newStorage array of characters to use as storage and set as new value
     * @see RawString#RawString(char[])
     */
    public void set(char[] newStorage) {
        this.storage = newStorage;
        this.startPos = 0;
        this.endPos = newStorage.length;
    }

    /**
     * Sets storage of this string to given array of characters and forms new value from
     * specified range of characters from the array.
     * <p>
     * After calling this method the instance will be in the same state, as if it were created
     * with <code>new RawString(newStorage, newStartPos, newLength)</code>.
     *
     * @param newStorage array of characters to use as storage
     * @param newStartPos
     *     index of first character in <code>characters</code> to use as this string's value
     * @param newLength
     *     number of characters from <code>characters</code> to use as this string's value
     * @throws IllegalArgumentException
     *     in case <code>newStartPos</code> or <code>newLength</code> is negative or their sum
     *     is grater than the length of <code>newStorage</code> array
     * @see RawString#RawString(char[], int, int)
     */
    public void set(char[] newStorage, int newStartPos, int newLength) {
        int newEndPos = newStartPos + newLength;
        if (!(newStartPos >= 0 && newLength >= 0 && newEndPos <= newStorage.length)) {
            throw new IllegalArgumentException(
                "Characters range specified with startPos=" + newStartPos
                + " and length=" + newLength + " is not valid or it doesn't fit in "
                + "[0, " + newStorage.length + "] range");
        }
        this.storage = newStorage;
        this.startPos = newStartPos;
        this.endPos = newEndPos;
    }

    /**
     * Sets storage of this string to storage of other {@link ReadableRawString} instance and
     * forms new value from value of that instance. After calling this method both strings will
     * be using the same storage and will have equal values.
     * <p>
     * After calling this method the instance will be in the same state, as if it were created
     * with <code>new RawString(other)</code>.
     *
     * @param other string to use storage and value from
     * @see RawString#RawString(ReadableRawString)
     */
    public void set(ReadableRawString other) {
        this.storage = other.storage;
        this.startPos = other.startPos;
        this.endPos = other.endPos;
    }

    /**
     * Sets a range of characters in this string's storage to use as this string's value.
     *
     * @param newStartPos
     *     new index of first character in this string's storage to use as this string's value
     * @param newLength
     *     number of characters from this string's storage to use as this string's value
     * @throws IllegalArgumentException
     *     in case <code>newStartPos</code> or <code>newLength</code> is negative or their sum
     *     is grater than the length of this string's storage
     */
    public void setRange(int newStartPos, int newLength) {
        int newEndPos = newStartPos + newLength;
        if (!(newStartPos >= 0 && newLength >= 0 && newEndPos <= storage.length)) {
            throw new IllegalArgumentException(
                "Characters range specified with startPos=" + newStartPos
                + " and length=" + newLength + " is not valid or it doesn't fit in "
                + "[0, " + storage.length + "] range");
        }
        startPos = newStartPos;
        endPos = newEndPos;
    }

    /**
     * Appends given character to current value of this string. The operation is performed in
     * two steps:
     * <ul>
     * <li>
     * the character is stored in this string's storage immediately after the last character
     * currently used as the string's value
     * </li>
     * <li>
     * index in this string's storage pointing immediately after the last character used as the
     * string's value gets increased (its value can be read using
     * {@link ReadableRawString#getEndPos()})
     * </li>
     * </ul>
     *
     * @param src character to append to current value
     * @throws RawStringOverflowException
     *     in case given character doesn't fit in the string's storage, i.e.
     *     <code>getFreeSpace() == 0</code>
     */
    public void append(char src) throws RawStringOverflowException {
        if (endPos + 1 > storage.length) {
            throw new RawStringOverflowException(src);
        }
        storage[endPos] = src;
        endPos++;
    }

    /**
     * Appends characters from given array to current value of this string. The operation is
     * performed in two steps:
     * <ul>
     * <li>
     * the characters are stored in this string's storage immediately after the last character
     * currently used as the string's value
     * </li>
     * <li>
     * index in this string's storage pointing immediately after the last character used as the
     * string's value gets increased (its value can be read using
     * {@link ReadableRawString#getEndPos()})
     * </li>
     * </ul>
     *
     * @param src array of characters to append to current value
     * @throws RawStringOverflowException
     *     in case given characters doesn't fit in the string's storage, i.e.
     *     <code>getFreeSpace() &lt; src.length</code>
     */
    public void append(char[] src) throws RawStringOverflowException {
        if (endPos + src.length > storage.length) {
            throw new RawStringOverflowException(src);
        }
        append(src, 0, src.length);
    }

    /**
     * Appends specified range of characters from given array to current value of this string.
     * The operation is performed in two steps:
     * <ul>
     * <li>
     * the characters are stored in this string's storage immediately after the last character
     * currently used as the string's value
     * </li>
     * <li>
     * index in this string's storage pointing immediately after the last character used as the
     * string's value gets increased (its value can be read using
     * {@link ReadableRawString#getEndPos()})
     * </li>
     * </ul>
     *
     * @param src array containing characters to append to current value
     * @param start index of first character in <code>src</code> to append
     * @param length number of characters to append
     * @throws IllegalArgumentException
     *     in case the following condition is not met: <code>start</code> &gt;= <code>0</code>
     *     &amp;&amp; <code>length</code> &gt;= <code>0</code> &amp;&amp;
     *     <code>start</code> + <code>length</code> &lt;= <code>src.length</code>
     * @throws RawStringOverflowException
     *     in case given characters doesn't fit in the string's storage, i.e.
     *     <code>getFreeSpace() &lt; length</code>
     */
    public void append(char[] src, int start, int length) throws RawStringOverflowException {
        if (!(start >= 0 && length >= 0 && start + length <= src.length)) {
            throw new IllegalArgumentException(
                "Parameters don't obey the rule:"
                + " start >= 0 && length >= 0 && start + length <= src.length"
                + " (start=" + start + ", length=" + length
                + ", src.length=" + src.length + ")");
        }
        if (endPos + length > storage.length) {
            throw new RawStringOverflowException(
                ArrayUtils.getCharSubArray(src, start, length));
        }
        System.arraycopy(src, start, storage, endPos, length);
        endPos += length;
    }

    /**
     * Appends value of given string to current value of this string. The operation is performed
     * in two steps:
     * <ul>
     * <li>
     * the characters that form value of given string are stored in this string's storage
     * immediately after the last character currently used as the string's value
     * </li>
     * <li>
     * index in this string's storage pointing immediately after the last character used as the
     * string's value gets increased (its value can be read using
     * {@link ReadableRawString#getEndPos()})
     * </li>
     * </ul>
     *
     * @param src string to append at the end of current value
     * @throws RawStringOverflowException
     *     in case characters from given string doesn't fit in this string's storage, i.e.
     *     <code>getFreeSpace() &lt; src.getLength()</code>
     */
    public void append(ReadableRawString src) throws RawStringOverflowException {
        int srcLen = src.getLength();
        if (endPos + srcLen > storage.length) {
            throw new RawStringOverflowException(
                ArrayUtils.getCharSubArray(src.getStorage(), src.getStartPos(), srcLen));
        }
        System.arraycopy(src.storage, src.startPos, storage, endPos, srcLen);
        endPos += srcLen;
    }

    /**
     * Appends specified range of characters from given string to current value of this string.
     * The operation is performed in two steps:
     * <ul>
     * <li>
     * the characters from value of given string are stored in this string's storage immediately
     * after the last character currently used as the string's value
     * </li>
     * <li>
     * index in this string's storage pointing immediately after the last character used as the
     * string's value gets increased (its value can be read using
     * {@link ReadableRawString#getEndPos()})
     * </li>
     * </ul>
     *
     * @param src string containing characters to append to current value
     * @param start
     *     index of first character in <code>src</code>'s value to append (it refers to index of
     *     character in string's value not in array of characters used internally as its
     *     storage)
     * @param length number of characters to append
     * @throws IllegalArgumentException
     *     in case the following condition is not met: <code>start</code> &gt;= <code>0</code>
     *     &amp;&amp; <code>length</code> &gt;= <code>0</code> &amp;&amp;
     *     <code>start</code> + <code>length</code> &lt;= <code>src.getLength()</code>
     * @throws RawStringOverflowException
     *     in case specified characters from given string don't fit in this string's storage,
     *     i.e. <code>getFreeSpace() &lt; length</code>
     */
    public void append(ReadableRawString src, int start, int length)
        throws RawStringOverflowException {
        if (!(start >= 0 && length >= 0 && start + length <= src.getLength())) {
            throw new IllegalArgumentException(
                "Parameters don't obey the rule:"
                + " start >= 0 && length >= 0 && start + length <= src.getLength()"
                + " (start=" + start + ", length=" + length
                + ", src.getLength()=" + src.getLength() + ")");
        }
        if (endPos + length > storage.length) {
            throw new RawStringOverflowException(
                ArrayUtils.getCharSubArray(src.storage, src.startPos + start, length));
        }
        System.arraycopy(
            src.storage, src.startPos + start, storage, endPos, length);
        endPos += length;
    }

    /**
     * Appends value of given string to current value of this string. The operation is performed
     * in two steps:
     * <ul>
     * <li>
     * the characters that form value of given string are stored in this string's storage
     * immediately after the last character currently used as the string's value
     * </li>
     * <li>
     * index in this string's storage pointing immediately after the last character used as the
     * string's value gets increased (its value can be read using
     * {@link ReadableRawString#getEndPos()})
     * </li>
     * </ul>
     *
     * @param src string to append at the end of current value
     * @throws RawStringOverflowException
     *     in case characters from given string doesn't fit in this string's storage, i.e.
     *     <code>getFreeSpace() &lt; src.length()</code>
     */
    public void append(String src) throws RawStringOverflowException {
        int srcLen = src.length();
        if (endPos + srcLen > storage.length) {
            throw new RawStringOverflowException(src.toCharArray());
        }
        src.getChars(0, srcLen, storage, endPos);
        endPos += srcLen;
    }

    /**
     * Appends specified range of characters from given string to current value of this string.
     * The operation is performed in two steps:
     * <ul>
     * <li>
     * the characters from value of given string are stored in this string's storage immediately
     * after the last character currently used as the string's value
     * </li>
     * <li>
     * index in this string's storage pointing immediately after the last character used as the
     * string's value gets increased (its value can be read using
     * {@link ReadableRawString#getEndPos()})
     * </li>
     * </ul>
     *
     * @param src string containing characters to append to current value
     * @param start index of first character in <code>src</code> to append
     * @param length number of characters to append
     * @throws IllegalArgumentException
     *     in case the following condition is not met: <code>start</code> &gt;= <code>0</code>
     *     &amp;&amp; <code>length</code> &gt;= <code>0</code> &amp;&amp;
     *     <code>start</code> + <code>length</code> &lt;= <code>src.length()</code>
     * @throws RawStringOverflowException
     *     in case specified characters from given string don't fit in this string's storage,
     *     i.e. <code>getFreeSpace() &lt; length</code>
     */
    public void append(String src, int start, int length) throws RawStringOverflowException {
        if (!(start >= 0 && length >= 0 && start + length <= src.length())) {
            throw new IllegalArgumentException(
                "Parameters don't obey the rule:"
                + " start >= 0 && length >= 0 && start + length <= src.length()"
                + " (start=" + start + ", length=" + length
                + ", src.length()=" + src.length() + ")");
        }
        if (endPos + length > storage.length) {
            throw new RawStringOverflowException(
                src.substring(start, start + length).toCharArray());
        }
        src.getChars(start, start + length, storage, endPos);
        endPos += length;
    }

    /**
     * Appends characters from given string buffer to current value of this string. The
     * operation is performed in two steps:
     * <ul>
     * <li>
     * the characters contained in given string buffer are stored in this string's storage
     * immediately after the last character currently used as the string's value
     * </li>
     * <li>
     * index in this string's storage pointing immediately after the last character used as the
     * string's value gets increased (its value can be read using
     * {@link ReadableRawString#getEndPos()})
     * </li>
     * </ul>
     *
     * @param src string buffer containing characters to append at the end of current value
     * @throws RawStringOverflowException
     *     in case characters from given string buffer don't fit in this string's storage, i.e.
     *     <code>getFreeSpace() &lt; src.length()</code>
     */
    public void append(StringBuffer src) throws RawStringOverflowException {
        int srcLen = src.length();
        if (endPos + srcLen > storage.length) {
            char[] srcChars = new char[src.length()];
            src.getChars(0, src.length(), srcChars, 0);
            throw new RawStringOverflowException(srcChars);
        }
        src.getChars(0, srcLen, storage, endPos);
        endPos += srcLen;
    }

    /**
     * Increases index in this string's storage pointing immediately after the last character
     * used as this string's value. Behaviour of this method is similar to the family of
     * <code>append</code> methods, except that it doesn't modify storage of the string. It lets
     * you to imitate the behaviour of <code>append</code> (including automatic checking of
     * storage overflow and throwing {@link RawStringOverflowException}) when you directly
     * manipulate storage of the string. For example, the following code is the equivalent of
     * <code>s.append("some string");</code>:
     * <blockquote><pre>
     * String as = "some string";
     * int oldPos = s.getEndPos();
     * s.extend(as.length());
     * as.getChars(0, as.length(), s.getStorage(), oldPos);
     * <pre><blockquote>
     *
     * @param extensionAmount
     *     number to add to index pointing after the last character used as this string's value
     * @throws IllegalArgumentException
     *     in case given <code>extensionAmount</code> is negative
     * @throws RawStringOverflowException
     *     in case given extension amount causes the increased string to point beyond the end of
     *     the storage
     */
    public void extend(int extensionAmount) {
        if (extensionAmount < 0) {
            throw new IllegalArgumentException(
                "Extension amount must not be negative"
                + "(extensionAmount = " + extensionAmount + ")");
        }
        if (endPos + extensionAmount > storage.length) {
            char[] emptyChars = new char[extensionAmount];
            Arrays.fill(emptyChars, ' ');
            throw new RawStringOverflowException(emptyChars);
        }
        endPos += extensionAmount;
    }

    /**
     * Returns number of unused components at the end of array used as this string's storage,
     * i.e. the difference between the storage's length and the index pointing immediately
     * after the last character in the storage that belongs to the string's value.
     *
     * @return number of unused characters at the end of this string's storage
     */
    public int getFreeSpace() {
        return storage.length - endPos;
    }


    /**
     * Exception thrown to indicate out of free space in storage of <code>RawString</code>
     * instance detected during appending some extra characters to <code>RawString</code>'s
     * value.
     *
     * @author Zbigniew Chyla (zbigniew@chyla.pl)
     */
    public final class RawStringOverflowException extends RuntimeException {

        /**
         * Characters that were to be added to value of string during appending operation that
         * caused an exception
         */
        private final char[] extraChars;

        /**
         * Constructs <code>RawStringOverflowException</code> instance that represents out of
         * storage situation detected during appending given characters to string's value.
         *
         * @param characters characters that were to be added to value of string
         */
        private RawStringOverflowException(char[] characters) {
            super(
                "Characters didn't fit in RawString's buffer:"
                + " \"" + new String(characters) + "\".");
            this.extraChars = characters;
        }

        /**
         * Constructs <code>RawStringOverflowException</code> instance that represents out of
         * storage situation detected during appending given character to string's value.
         *
         * @param character character that was to be added to value of string
         */
        public RawStringOverflowException(char character) {
            super("Character didn't fit in RawString's buffer: \"" + character + "\".");
            this.extraChars = new char[] {character};
        }

        /**
         * Returns <code>RawString</code> instance on which appending operation (that caused
         * throwing <code>RawStringOverflowException</code>) was performed.
         *
         * @return <code>RawString</code> instance that got overflowed during appending
         */
        public RawString getOverflowedString() {
            return RawString.this;
        }

        /**
         * Returns characters that were to be added to value of string during appending
         * operation that caused an exception
         *
         * @return
         *     characters that were not appended to value of string because of missing free
         *     space in its storage
         */
        public char[] getExtraCharacters() {
            return extraChars;
        }

    }

}
