/*
 * Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License version 2.1 as published by the Free Software
 * Foundation.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301  USA
 */
package pl.chyla.wyjec.utils;


/**
 * Very simple (and featureless) abstract replacement for standard String class. It gives direct
 * access to underlying array of characters (called <emph>storage</emph> in this context) thus
 * allowing for faster access to individual characters.
 * <p>
 * ReadableRawString instances are read-only in terms of API as none of its methods modify its
 * content. However it's possible to replace the string's characters by direct manipulation
 * on underlying array. Such operations are strongly discouraged (unfortunately Java language
 * doesn't have "const" modifier so we can't force this limitation without hiding array of
 * characters completely).
 * <p>
 * Presented approach allows for sharing storage between many <code>ReadableRawString</code>
 * instances, even if they have different values, e.g. we may have instances <code>A</code>
 * and <code>B</code>, both using the same storage <code>{'a', 'b', 'c', 'd'}</code>, but
 * <code>A</code> may use characters with indices 0-2 for its value ("abc") and <code>B</code>
 * may use characters with indices 2-3 ("cd"). Sharing storage allows for efficient generating
 * of substrings and executing some other operations, but it should be used very carefully
 * because modifications of characters inside one of the strings may cause (unexpected) changes
 * of values of many others.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public abstract class ReadableRawString implements Comparable {

    /**
     * Value used as number of bits to shift intermediate value at every subsequent character
     * position when calculating hash for a string.
     */
    private static final int HASH_CODE_CHAR_SHIFT = 5;

    /**
     * Array of characters used as storage for the string
     */
    protected char[] storage;

    /**
     * Index of first character in <code>characters</code> used as the string's value
     */
    protected int startPos;

    /**
     * Index in <code>storage</code> pointing immediately after the last character used as the
     * string's value
     */
    protected int endPos;

    /**
     * Returns current storage of the string, i.e. array of characters containing its value
     * (potentially only as part of the array).
     *
     * @return array of characters used as storage for the string
     */
    public final char[] getStorage() {
        return storage;
    }

    /**
     * Returns current length of the string, i.e. number of characters forming its value.
     *
     * @return length of the string (which may be smaller that the length of its storage)
     */
    public final int getLength() {
        return endPos - startPos;
    }

    /**
     * Returns index in the strings's storage (as returned by {@link #getStorage()}) that
     * points to the first character used as the string's value.
     *
     * @return index of first character used as the string's value in the string's storage
     */
    public final int getStartPos() {
        return startPos;
    }

    /**
     * Returns index in the strings's storage (as returned by {@link #getStorage()}) that
     * points immediately after the last character used as the string's value.
     *
     * @return
     *     index pointing immediately after the last character used as the string's value in the
     *     string's storage
     */
    public final int getEndPos() {
        return endPos;
    }

    /**
     * Generates representation of the string as <code>String</code> instance.
     *
     * @return value of the string as <code>String</code> instance
     */
    public final String toString() {
        return String.copyValueOf(storage, startPos, getLength());
    }

    /**
     * Checks whether the string is equal to another one, comparing their values (not
     * necessarily the whole storages).
     *
     * @param other string to compare this string to
     * @return <code>true</code> if both strings are equal, <code>false</code> otherwise
     */
    public final boolean equals(ReadableRawString other) {
        if (other == null) {
            return false;
        }
        if (getLength() != other.getLength()) {
            return false;
        }
        if (this == other) {
            return true;
        }
        char[] chars = storage;
        char[] otherChars = other.storage;
        int i = startPos;
        int iLimit = endPos;
        int j = other.startPos;
        while (i < iLimit) {
            if (chars[i] != otherChars[j]) {
                return false;
            }
            i++;
            j++;
        }
        return true;
    }

    /**
     * Compares the string to another one for order, performing lexicographical comparision.
     *
     * @param other string to compare this string to
     * @return
     *     negative number, zero or positive number if value of the string is less, equal or
     *     greater than value of <code>other</code>
     */
    public final int compareTo(ReadableRawString other) {
        int len = getLength();
        int otherLen = other.getLength();
        char[] chars = storage;
        char[] otherChars = other.storage;
        int i = startPos;
        int j = other.startPos;
        int iLimit = startPos + Math.min(len, otherLen);
        while (i < iLimit && chars[i] == otherChars[j]) {
            i++;
            j++;
        }
        int result;
        if (i < iLimit) {
            if (chars[i] < otherChars[j]) {
                result = -1;
            } else {
                result = 1;
            }
        } else {
            result = IntUtils.compare(len, otherLen);
        }
        return result;
    }

    /**
     * Searches forward for given string in contents of this string, starting from given
     * position. If given string (to search for) is empty, this method always returns specified
     * start position (<code>findStartPos</code>).
     *
     * @param subString string to search inside this string
     * @param findStartPos position inside this string to start searching from
     * @return
     *     the lowest index in this string not less than <code>findStartPos</code>, at which
     *     given substring is found or <code>-1</code> if this string doesn't contain given
     *     substring or it's located before <code>findStartPos</code>
     * @throws IllegalArgumentException
     *     in case <code>findStartPos</code> is negative or greater than current length of this
     *     string
     */
    public final int find(ReadableRawString subString, int findStartPos) {
        if (findStartPos < 0 || findStartPos > getLength()) {
            throw new IllegalArgumentException(
                "Invalid search start position for this string: " + findStartPos);
        }
        int subStrStartPos = subString.startPos;
        int subStrLen = subString.getLength();
        char[] chars = storage;
        char[] subStrChars = subString.storage;
        int posLimit = endPos - (subStrLen - 1);
        int foundPos = -1;
    iterPositions:
        for (int i = startPos + findStartPos; i < posLimit; i++) {
            for (int j = 0; j < subStrLen; j++) {
                if (subStrChars[subStrStartPos + j] != chars[i + j]) {
                    continue iterPositions;
                }
            }
            foundPos = i - startPos;
            break;
        }
        return foundPos;
    }

    /**
     * Searches forward for given character in contents of this string, starting from given
     * position.
     *
     * @param ch character to search inside this string
     * @param findStartPos position inside this string to start searching from
     * @return
     *     the lowest index in this string not less than <code>findStartPos</code>, at which
     *     given character is found or <code>-1</code> if this string doesn't contain given
     *     substring or it's located before <code>findStartPos</code>
     * @throws IllegalArgumentException
     *     in case <code>findStartPos</code> is negative or greater than current length of this
     *     string
     */
    public final int find(char ch, int findStartPos) {
        if (findStartPos < 0 || findStartPos > getLength()) {
            throw new IllegalArgumentException(
                "Invalid search start position for this string: " + findStartPos);
        }
        char[] chars = storage;
        int iLimit = endPos;
        int foundPos = -1;
        for (int i = startPos + findStartPos; i < iLimit; i++) {
            if (chars[i] == ch) {
                foundPos = i - startPos;
                break;
            }
        }
        return foundPos;
    }

    /**
     * Compares the string to another one for order, performing lexicographical comparision.
     *
     * @param other string to compare this string to
     * @return
     *     negative number, zero or positive number if value of the string is less, equal or
     *     greater than value of <code>other</code>
     * @throws ClassCastException
     *     in case <code>other</code> is not <code>ReadableRawString</code> instance
     */
    public final int compareTo(Object other) {
        return compareTo((ReadableRawString) other);
    }

    /**
     * Checks whether the string is equal to another object.
     *
     * @param other object to compare this string to
     * @return
     *     <code>true</code> if <code>other</code> is <code>ReadableRawString</code> instance
     *     with value equal to the string, <code>false</code> otherwise
     */
    public final boolean equals(Object other) {
        if (other instanceof ReadableRawString) {
            return equals((ReadableRawString) other);
        } else {
            return false;
        }
    }

    /**
     * Returns a hash code value for the string, using only its value. The method will always
     * return the same value for two different <code>ReadableRawString</code> instances with
     * equal values.
     *
     * @return returns a hash code value for the string
     */
    public final int hashCode() {
        int hash = 0;
        for (int i = startPos; i < endPos; i++) {
            hash = (hash << HASH_CODE_CHAR_SHIFT) - hash + storage[i];
        }
        return hash;
    }

}
